#!/usr/bin/env python
# Copyright (c) 2017 The WebRTC project authors. All Rights Reserved.
#
# Use of this source code is governed by a BSD-style license
# that can be found in the LICENSE file in the root of the source
# tree. An additional intellectual property rights grant can be found
# in the file PATENTS.  All contributing project authors may
# be found in the AUTHORS file in the root of the source tree.
"""Export the scores computed by the apm_quality_assessment.py script into an
   HTML file.
"""

import logging
import os
import sys

import quality_assessment.collect_data as collect_data
import quality_assessment.export as export


def _BuildOutputFilename(filename_suffix):
    """Builds the filename for the exported file.

  Args:
    filename_suffix: suffix for the output file name.

  Returns:
    A string.
  """
    if filename_suffix is None:
        return 'results.html'
    return 'results-{}.html'.format(filename_suffix)


def main():
    # Init.
    logging.basicConfig(
        level=logging.DEBUG)  # TODO(alessio): INFO once debugged.
    parser = collect_data.InstanceArgumentsParser()
    parser.add_argument('-f',
                        '--filename_suffix',
                        help=('suffix of the exported file'))
    parser.description = ('Exports pre-computed APM module quality assessment '
                          'results into HTML tables')
    args = parser.parse_args()

    # Get the scores.
    src_path = collect_data.ConstructSrcPath(args)
    logging.debug(src_path)
    scores_data_frame = collect_data.FindScores(src_path, args)

    # Export.
    output_filepath = os.path.join(args.output_dir,
                                   _BuildOutputFilename(args.filename_suffix))
    exporter = export.HtmlExport(output_filepath)
    exporter.Export(scores_data_frame)

    logging.info('output file successfully written in %s', output_filepath)
    sys.exit(0)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# Copyright (c) 2017 The WebRTC project authors. All Rights Reserved.
#
# Use of this source code is governed by a BSD-style license
# that can be found in the LICENSE file in the root of the source
# tree. An additional intellectual property rights grant can be found
# in the file PATENTS.  All contributing project authors may
# be found in the AUTHORS file in the root of the source tree.
"""Generate .json files with which the APM module can be tested using the
   apm_quality_assessment.py script and audioproc_f as APM simulator.
"""

import logging
import os

import quality_assessment.data_access as data_access

OUTPUT_PATH = os.path.abspath('apm_configs')


def _GenerateDefaultOverridden(config_override):
    """Generates one or more APM overriden configurations.

  For each item in config_override, it overrides the default configuration and
  writes a new APM configuration file.

  The default settings are loaded via "-all_default".
  Check "src/modules/audio_processing/test/audioproc_float.cc" and search
  for "if (FLAG_all_default) {".

  For instance, in 55eb6d621489730084927868fed195d3645a9ec9 the default is this:
  settings.use_aec = rtc::Optional<bool>(true);
  settings.use_aecm = rtc::Optional<bool>(false);
  settings.use_agc = rtc::Optional<bool>(true);
  settings.use_bf = rtc::Optional<bool>(false);
  settings.use_ed = rtc::Optional<bool>(false);
  settings.use_hpf = rtc::Optional<bool>(true);
  settings.use_le = rtc::Optional<bool>(true);
  settings.use_ns = rtc::Optional<bool>(true);
  settings.use_ts = rtc::Optional<bool>(true);
  settings.use_vad = rtc::Optional<bool>(true);

  Args:
    config_override: dict of APM configuration file names as keys; the values
                     are dict instances encoding the audioproc_f flags.
  """
    for config_filename in config_override:
        config = config_override[config_filename]
        config['-all_default'] = None

        config_filepath = os.path.join(
            OUTPUT_PATH, 'default-{}.json'.format(config_filename))
        logging.debug('config file <%s> | %s', config_filepath, config)

        data_access.AudioProcConfigFile.Save(config_filepath, config)
        logging.info('config file created: <%s>', config_filepath)


def _GenerateAllDefaultButOne():
    """Disables the flags enabled by default one-by-one.
  """
    config_sets = {
        'no_AEC': {
            '-aec': 0,
        },
        'no_AGC': {
            '-agc': 0,
        },
        'no_HP_filter': {
            '-hpf': 0,
        },
        'no_level_estimator': {
            '-le': 0,
        },
        'no_noise_suppressor': {
            '-ns': 0,
        },
        'no_transient_suppressor': {
            '-ts': 0,
        },
        'no_vad': {
            '-vad': 0,
        },
    }
    _GenerateDefaultOverridden(config_sets)


def _GenerateAllDefaultPlusOne():
    """Enables the flags disabled by default one-by-one.
  """
    config_sets = {
        'with_AECM': {
            '-aec': 0,
            '-aecm': 1,
        },  # AEC and AECM are exclusive.
        'with_AGC_limiter': {
            '-agc_limiter': 1,
        },
        'with_AEC_delay_agnostic': {
            '-delay_agnostic': 1,
        },
        'with_drift_compensation': {
            '-drift_compensation': 1,
        },
        'with_residual_echo_detector': {
            '-ed': 1,
        },
        'with_AEC_extended_filter': {
            '-extended_filter': 1,
        },
        'with_LC': {
            '-lc': 1,
        },
        'with_refined_adaptive_filter': {
            '-refined_adaptive_filter': 1,
        },
    }
    _GenerateDefaultOverridden(config_sets)


def main():
    logging.basicConfig(level=logging.INFO)
    _GenerateAllDefaultPlusOne()
    _GenerateAllDefaultButOne()


if __name__ == '__main__':
    main()



# Copyright (c) 2017 The WebRTC project authors. All Rights Reserved.
#
# Use of this source code is governed by a BSD-style license
# that can be found in the LICENSE file in the root of the source
# tree. An additional intellectual property rights grant can be found
# in the file PATENTS.  All contributing project authors may
# be found in the AUTHORS file in the root of the source tree.
"""Extraction of annotations from audio files.
"""

from __future__ import division
import logging
import os
import shutil
import struct
import subprocess
import sys
import tempfile

try:
    import numpy as np
except ImportError:
    logging.critical('Cannot import the third-party Python package numpy')
    sys.exit(1)

from . import external_vad
from . import exceptions
from . import signal_processing


class AudioAnnotationsExtractor(object):
    """Extracts annotations from audio files.
  """

    class VadType(object):
        ENERGY_THRESHOLD = 1  # TODO(alessiob): Consider switching to P56 standard.
        WEBRTC_COMMON_AUDIO = 2  # common_audio/vad/include/vad.h
        WEBRTC_APM = 4  # modules/audio_processing/vad/vad.h

        def __init__(self, value):
            if (not isinstance(value, int)) or not 0 <= value <= 7:
                raise exceptions.InitializationException('Invalid vad type: ' +
                                                         value)
            self._value = value

        def Contains(self, vad_type):
            return self._value | vad_type == self._value

        def __str__(self):
            vads = []
            if self.Contains(self.ENERGY_THRESHOLD):
                vads.append("energy")
            if self.Contains(self.WEBRTC_COMMON_AUDIO):
                vads.append("common_audio")
            if self.Contains(self.WEBRTC_APM):
                vads.append("apm")
            return "VadType({})".format(", ".join(vads))

    _OUTPUT_FILENAME_TEMPLATE = '{}annotations.npz'

    # Level estimation params.
    _ONE_DB_REDUCTION = np.power(10.0, -1.0 / 20.0)
    _LEVEL_FRAME_SIZE_MS = 1.0
    # The time constants in ms indicate the time it takes for the level estimate
    # to go down/up by 1 db if the signal is zero.
    _LEVEL_ATTACK_MS = 5.0
    _LEVEL_DECAY_MS = 20.0

    # VAD params.
    _VAD_THRESHOLD = 1
    _VAD_WEBRTC_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                    os.pardir, os.pardir)
    _VAD_WEBRTC_COMMON_AUDIO_PATH = os.path.join(_VAD_WEBRTC_PATH, 'vad')

    _VAD_WEBRTC_APM_PATH = os.path.join(_VAD_WEBRTC_PATH, 'apm_vad')

    def __init__(self, vad_type, external_vads=None):
        self._signal = None
        self._level = None
        self._level_frame_size = None
        self._common_audio_vad = None
        self._energy_vad = None
        self._apm_vad_probs = None
        self._apm_vad_rms = None
        self._vad_frame_size = None
        self._vad_frame_size_ms = None
        self._c_attack = None
        self._c_decay = None

        self._vad_type = self.VadType(vad_type)
        logging.info('VADs used for annotations: ' + str(self._vad_type))

        if external_vads is None:
            external_vads = {}
        self._external_vads = external_vads

        assert len(self._external_vads) == len(external_vads), (
            'The external VAD names must be unique.')
        for vad in external_vads.values():
            if not isinstance(vad, external_vad.ExternalVad):
                raise exceptions.InitializationException('Invalid vad type: ' +
                                                         str(type(vad)))
            logging.info('External VAD used for annotation: ' + str(vad.name))

        assert os.path.exists(self._VAD_WEBRTC_COMMON_AUDIO_PATH), \
          self._VAD_WEBRTC_COMMON_AUDIO_PATH
        assert os.path.exists(self._VAD_WEBRTC_APM_PATH), \
          self._VAD_WEBRTC_APM_PATH

    @classmethod
    def GetOutputFileNameTemplate(cls):
        return cls._OUTPUT_FILENAME_TEMPLATE

    def GetLevel(self):
        return self._level

    def GetLevelFrameSize(self):
        return self._level_frame_size

    @classmethod
    def GetLevelFrameSizeMs(cls):
        return cls._LEVEL_FRAME_SIZE_MS

    def GetVadOutput(self, vad_type):
        if vad_type == self.VadType.ENERGY_THRESHOLD:
            return self._energy_vad
        elif vad_type == self.VadType.WEBRTC_COMMON_AUDIO:
            return self._common_audio_vad
        elif vad_type == self.VadType.WEBRTC_APM:
            return (self._apm_vad_probs, self._apm_vad_rms)
        else:
            raise exceptions.InitializationException('Invalid vad type: ' +
                                                     vad_type)

    def GetVadFrameSize(self):
        return self._vad_frame_size

    def GetVadFrameSizeMs(self):
        return self._vad_frame_size_ms

    def Extract(self, filepath):
        # Load signal.
        self._signal = signal_processing.SignalProcessingUtils.LoadWav(
            filepath)
        if self._signal.channels != 1:
            raise NotImplementedError(
                'Multiple-channel annotations not implemented')

        # Level estimation params.
        self._level_frame_size = int(self._signal.frame_rate / 1000 *
                                     (self._LEVEL_FRAME_SIZE_MS))
        self._c_attack = 0.0 if self._LEVEL_ATTACK_MS == 0 else (
            self._ONE_DB_REDUCTION**(self._LEVEL_FRAME_SIZE_MS /
                                     self._LEVEL_ATTACK_MS))
        self._c_decay = 0.0 if self._LEVEL_DECAY_MS == 0 else (
            self._ONE_DB_REDUCTION**(self._LEVEL_FRAME_SIZE_MS /
                                     self._LEVEL_DECAY_MS))

        # Compute level.
        self._LevelEstimation()

        # Ideal VAD output, it requires clean speech with high SNR as input.
        if self._vad_type.Contains(self.VadType.ENERGY_THRESHOLD):
            # Naive VAD based on level thresholding.
            vad_threshold = np.percentile(self._level, self._VAD_THRESHOLD)
            self._energy_vad = np.uint8(self._level > vad_threshold)
            self._vad_frame_size = self._level_frame_size
            self._vad_frame_size_ms = self._LEVEL_FRAME_SIZE_MS
        if self._vad_type.Contains(self.VadType.WEBRTC_COMMON_AUDIO):
            # WebRTC common_audio/ VAD.
            self._RunWebRtcCommonAudioVad(filepath, self._signal.frame_rate)
        if self._vad_type.Contains(self.VadType.WEBRTC_APM):
            # WebRTC modules/audio_processing/ VAD.
            self._RunWebRtcApmVad(filepath)
        for extvad_name in self._external_vads:
            self._external_vads[extvad_name].Run(filepath)

    def Save(self, output_path, annotation_name=""):
        ext_kwargs = {
            'extvad_conf-' + ext_vad:
            self._external_vads[ext_vad].GetVadOutput()
            for ext_vad in self._external_vads
        }
        np.savez_compressed(file=os.path.join(
            output_path,
            self.GetOutputFileNameTemplate().format(annotation_name)),
                            level=self._level,
                            level_frame_size=self._level_frame_size,
                            level_frame_size_ms=self._LEVEL_FRAME_SIZE_MS,
                            vad_output=self._common_audio_vad,
                            vad_energy_output=self._energy_vad,
                            vad_frame_size=self._vad_frame_size,
                            vad_frame_size_ms=self._vad_frame_size_ms,
                            vad_probs=self._apm_vad_probs,
                            vad_rms=self._apm_vad_rms,
                            **ext_kwargs)

    def _LevelEstimation(self):
        # Read samples.
        samples = signal_processing.SignalProcessingUtils.AudioSegmentToRawData(
            self._signal).astype(np.float32) / 32768.0
        num_frames = len(samples) // self._level_frame_size
        num_samples = num_frames * self._level_frame_size

        # Envelope.
        self._level = np.max(np.reshape(np.abs(samples[:num_samples]),
                                        (num_frames, self._level_frame_size)),
                             axis=1)
        assert len(self._level) == num_frames

        # Envelope smoothing.
        smooth = lambda curr, prev, k: (1 - k) * curr + k * prev
        self._level[0] = smooth(self._level[0], 0.0, self._c_attack)
        for i in range(1, num_frames):
            self._level[i] = smooth(
                self._level[i], self._level[i - 1], self._c_attack if
                (self._level[i] > self._level[i - 1]) else self._c_decay)

    def _RunWebRtcCommonAudioVad(self, wav_file_path, sample_rate):
        self._common_audio_vad = None
        self._vad_frame_size = None

        # Create temporary output path.
        tmp_path = tempfile.mkdtemp()
        output_file_path = os.path.join(
            tmp_path,
            os.path.split(wav_file_path)[1] + '_vad.tmp')

        # Call WebRTC VAD.
        try:
            subprocess.call([
                self._VAD_WEBRTC_COMMON_AUDIO_PATH, '-i', wav_file_path, '-o',
                output_file_path
            ],
                            cwd=self._VAD_WEBRTC_PATH)

            # Read bytes.
            with open(output_file_path, 'rb') as f:
                raw_data = f.read()

            # Parse side information.
            self._vad_frame_size_ms = struct.unpack('B', raw_data[0])[0]
            self._vad_frame_size = self._vad_frame_size_ms * sample_rate / 1000
            assert self._vad_frame_size_ms in [10, 20, 30]
            extra_bits = struct.unpack('B', raw_data[-1])[0]
            assert 0 <= extra_bits <= 8

            # Init VAD vector.
            num_bytes = len(raw_data)
            num_frames = 8 * (num_bytes -
                              2) - extra_bits  # 8 frames for each byte.
            self._common_audio_vad = np.zeros(num_frames, np.uint8)

            # Read VAD decisions.
            for i, byte in enumerate(raw_data[1:-1]):
                byte = struct.unpack('B', byte)[0]
                for j in range(8 if i < num_bytes - 3 else (8 - extra_bits)):
                    self._common_audio_vad[i * 8 + j] = int(byte & 1)
                    byte = byte >> 1
        except Exception as e:
            logging.error('Error while running the WebRTC VAD (' + e.message +
                          ')')
        finally:
            if os.path.exists(tmp_path):
                shutil.rmtree(tmp_path)

    def _RunWebRtcApmVad(self, wav_file_path):
        # Create temporary output path.
        tmp_path = tempfile.mkdtemp()
        output_file_path_probs = os.path.join(
            tmp_path,
            os.path.split(wav_file_path)[1] + '_vad_probs.tmp')
        output_file_path_rms = os.path.join(
            tmp_path,
            os.path.split(wav_file_path)[1] + '_vad_rms.tmp')

        # Call WebRTC VAD.
        try:
            subprocess.call([
                self._VAD_WEBRTC_APM_PATH, '-i', wav_file_path, '-o_probs',
                output_file_path_probs, '-o_rms', output_file_path_rms
            ],
                            cwd=self._VAD_WEBRTC_PATH)

            # Parse annotations.
            self._apm_vad_probs = np.fromfile(output_file_path_probs,
                                              np.double)
            self._apm_vad_rms = np.fromfile(output_file_path_rms, np.double)
            assert len(self._apm_vad_rms) == len(self._apm_vad_probs)

        except Exception as e:
            logging.error('Error while running the WebRTC APM VAD (' +
                          e.message + ')')
        finally:
            if os.path.exists(tmp_path):
                shutil.rmtree(tmp_path)



# Copyright 2017 The CRC32C Authors. All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.
"""YouCompleteMe configuration that interprets a .clang_complete file.

This module implementes the YouCompleteMe configuration API documented at:
https://github.com/ycm-core/ycmd#ycm_extra_confpy-specification

The implementation loads and processes a .clang_complete file, documented at:
https://github.com/xavierd/clang_complete/blob/master/README.md
"""

import os

# Flags added to the list in .clang_complete.
BASE_FLAGS = [
    '-Werror',  # Unlike clang_complete, YCM can also be used as a linter.
    '-DUSE_CLANG_COMPLETER',  # YCM needs this.
    '-xc++',  # YCM needs this to avoid compiling headers as C code.
]

# Clang flags that take in paths.
# See https://clang.llvm.org/docs/ClangCommandLineReference.html
PATH_FLAGS = [
    '-isystem',
    '-I',
    '-iquote',
    '--sysroot='
]


def DirectoryOfThisScript():
  """Returns the absolute path to the directory containing this script."""
  return os.path.dirname(os.path.abspath(__file__))


def MakeRelativePathsInFlagsAbsolute(flags, build_root):
  """Expands relative paths in a list of Clang command-line flags.

  Args:
    flags: The list of flags passed to Clang.
    build_root: The current directory when running the Clang compiler. Should be
        an absolute path.

  Returns:
    A list of flags with relative paths replaced by absolute paths.
  """
  new_flags = []
  make_next_absolute = False
  for flag in flags:
    new_flag = flag

    if make_next_absolute:
      make_next_absolute = False
      if not flag.startswith('/'):
        new_flag = os.path.join(build_root, flag)

    for path_flag in PATH_FLAGS:
      if flag == path_flag:
        make_next_absolute = True
        break

      if flag.startswith(path_flag):
        path = flag[len(path_flag):]
        new_flag = path_flag + os.path.join(build_root, path)
        break

    if new_flag:
      new_flags.append(new_flag)
  return new_flags


def FindNearest(target, path, build_root):
  """Looks for a file with a specific name closest to a project path.

  This is similar to the logic used by a version-control system (like git) to
  find its configuration directory (.git) based on the current directory when a
  command is invoked.

  Args:
    target: The file name to search for.
    path: The directory where the search starts. The search will explore the
        given directory's ascendants using the parent relationship. Should be an
        absolute path.
    build_root: A directory that acts as a fence for the search. If the search
        reaches this directory, it will not advance to its parent. Should be an
        absolute path.

  Returns:
    The path to a file with the desired name. None if the search failed.
  """
  candidate = os.path.join(path, target)
  if os.path.isfile(candidate):
    return candidate

  if path == build_root:
    return None

  parent = os.path.dirname(path)
  if parent == path:
    return None

  return FindNearest(target, parent, build_root)


def FlagsForClangComplete(file_path, build_root):
  """Reads the .clang_complete flags for a source file.

  Args:
    file_path: The path to the source file. Should be inside the project. Used
      to locate the relevant .clang_complete file.
    build_root: The current directory when running the Clang compiler for this
        file. Should be an absolute path.

  Returns:
    A list of strings, where each element is a Clang command-line flag.
  """
  clang_complete_path = FindNearest('.clang_complete', file_path, build_root)
  if clang_complete_path is None:
    return None
  clang_complete_flags = open(clang_complete_path, 'r').read().splitlines()
  return clang_complete_flags


def FlagsForFile(filename, **kwargs):
  """Implements the YouCompleteMe API."""

  # kwargs can be used to pass 'client_data' to the YCM configuration. This
  # configuration script does not need any extra information, so
  # pylint: disable=unused-argument

  build_root = DirectoryOfThisScript()
  file_path = os.path.realpath(filename)

  flags = BASE_FLAGS
  clang_flags = FlagsForClangComplete(file_path, build_root)
  if clang_flags:
    flags += clang_flags

  final_flags = MakeRelativePathsInFlagsAbsolute(flags, build_root)

  return {'flags': final_flags}



# Copyright (c) 2017 The WebRTC project authors. All Rights Reserved.
#
# Use of this source code is governed by a BSD-style license
# that can be found in the LICENSE file in the root of the source
# tree. An additional intellectual property rights grant can be found
# in the file PATENTS.  All contributing project authors may
# be found in the AUTHORS file in the root of the source tree.
"""Evaluator of the APM module.
"""

import logging


class ApmModuleEvaluator(object):
    """APM evaluator class.
  """

    def __init__(self):
        pass

    @classmethod
    def Run(cls, evaluation_score_workers, apm_input_metadata,
            apm_output_filepath, reference_input_filepath,
            render_input_filepath, output_path):
        """Runs the evaluation.

    Iterates over the given evaluation score workers.

    Args:
      evaluation_score_workers: list of EvaluationScore instances.
      apm_input_metadata: dictionary with metadata of the APM input.
      apm_output_filepath: path to the audio track file with the APM output.
      reference_input_filepath: path to the reference audio track file.
      output_path: output path.

    Returns:
      A dict of evaluation score name and score pairs.
    """
        # Init.
        scores = {}

        for evaluation_score_worker in evaluation_score_workers:
            logging.info('   computing <%s> score',
                         evaluation_score_worker.NAME)
            evaluation_score_worker.SetInputSignalMetadata(apm_input_metadata)
            evaluation_score_worker.SetReferenceSignalFilepath(
                reference_input_filepath)
            evaluation_score_worker.SetTestedSignalFilepath(
                apm_output_filepath)
            evaluation_score_worker.SetRenderSignalFilepath(
                render_input_filepath)

            evaluation_score_worker.Run(output_path)
            scores[
                evaluation_score_worker.NAME] = evaluation_score_worker.score

        return scores



#  Copyright (c) 2017 The WebRTC project authors. All Rights Reserved.
#
#  Use of this source code is governed by a BSD-style license
#  that can be found in the LICENSE file in the root of the source
#  tree. An additional intellectual property rights grant can be found
#  in the file PATENTS.  All contributing project authors may
#  be found in the AUTHORS file in the root of the source tree.
"""Plots statistics from WebRTC integration test logs.

Usage: $ python plot_webrtc_test_logs.py filename.txt
"""

import numpy
import sys
import re

import matplotlib.pyplot as plt

# Log events.
EVENT_START = 'RUN      ] CodecSettings/VideoCodecTestParameterized.'
EVENT_END = 'OK ] CodecSettings/VideoCodecTestParameterized.'

# Metrics to plot, tuple: (name to parse in file, label to use when plotting).
WIDTH = ('width', 'width')
HEIGHT = ('height', 'height')
FILENAME = ('filename', 'clip')
CODEC_TYPE = ('codec_type', 'Codec')
ENCODER_IMPLEMENTATION_NAME = ('enc_impl_name', 'enc name')
DECODER_IMPLEMENTATION_NAME = ('dec_impl_name', 'dec name')
CODEC_IMPLEMENTATION_NAME = ('codec_impl_name', 'codec name')
CORES = ('num_cores', 'CPU cores used')
DENOISING = ('denoising', 'denoising')
RESILIENCE = ('resilience', 'resilience')
ERROR_CONCEALMENT = ('error_concealment', 'error concealment')
CPU_USAGE = ('cpu_usage_percent', 'CPU usage (%)')
BITRATE = ('target_bitrate_kbps', 'target bitrate (kbps)')
FRAMERATE = ('input_framerate_fps', 'fps')
QP = ('avg_qp', 'QP avg')
PSNR = ('avg_psnr', 'PSNR (dB)')
SSIM = ('avg_ssim', 'SSIM')
ENC_BITRATE = ('bitrate_kbps', 'encoded bitrate (kbps)')
NUM_FRAMES = ('num_input_frames', 'num frames')
NUM_DROPPED_FRAMES = ('num_dropped_frames', 'num dropped frames')
TIME_TO_TARGET = ('time_to_reach_target_bitrate_sec',
                  'time to reach target rate (sec)')
ENCODE_SPEED_FPS = ('enc_speed_fps', 'encode speed (fps)')
DECODE_SPEED_FPS = ('dec_speed_fps', 'decode speed (fps)')
AVG_KEY_FRAME_SIZE = ('avg_key_frame_size_bytes', 'avg key frame size (bytes)')
AVG_DELTA_FRAME_SIZE = ('avg_delta_frame_size_bytes',
                        'avg delta frame size (bytes)')

# Settings.
SETTINGS = [
    WIDTH,
    HEIGHT,
    FILENAME,
    NUM_FRAMES,
]

# Settings, options for x-axis.
X_SETTINGS = [
    CORES,
    FRAMERATE,
    DENOISING,
    RESILIENCE,
    ERROR_CONCEALMENT,
    BITRATE,  # TODO(asapersson): Needs to be last.
]

# Settings, options for subplots.
SUBPLOT_SETTINGS = [
    CODEC_TYPE,
    ENCODER_IMPLEMENTATION_NAME,
    DECODER_IMPLEMENTATION_NAME,
    CODEC_IMPLEMENTATION_NAME,
] + X_SETTINGS

# Results.
RESULTS = [
    PSNR,
    SSIM,
    ENC_BITRATE,
    NUM_DROPPED_FRAMES,
    TIME_TO_TARGET,
    ENCODE_SPEED_FPS,
    DECODE_SPEED_FPS,
    QP,
    CPU_USAGE,
    AVG_KEY_FRAME_SIZE,
    AVG_DELTA_FRAME_SIZE,
]

METRICS_TO_PARSE = SETTINGS + SUBPLOT_SETTINGS + RESULTS

Y_METRICS = [res[1] for res in RESULTS]

# Parameters for plotting.
FIG_SIZE_SCALE_FACTOR_X = 1.6
FIG_SIZE_SCALE_FACTOR_Y = 1.8
GRID_COLOR = [0.45, 0.45, 0.45]


def ParseSetting(filename, setting):
    """Parses setting from file.

  Args:
    filename: The name of the file.
    setting: Name of setting to parse (e.g. width).

  Returns:
    A list holding parsed settings, e.g. ['width: 128.0', 'width: 160.0'] """

    settings = []

    settings_file = open(filename)
    while True:
        line = settings_file.readline()
        if not line:
            break
        if re.search(r'%s' % EVENT_START, line):
            # Parse event.
            parsed = {}
            while True:
                line = settings_file.readline()
                if not line:
                    break
                if re.search(r'%s' % EVENT_END, line):
                    # Add parsed setting to list.
                    if setting in parsed:
                        s = setting + ': ' + str(parsed[setting])
                        if s not in settings:
                            settings.append(s)
                    break

                TryFindMetric(parsed, line)

    settings_file.close()
    return settings


def ParseMetrics(filename, setting1, setting2):
    """Parses metrics from file.

  Args:
    filename: The name of the file.
    setting1: First setting for sorting metrics (e.g. width).
    setting2: Second setting for sorting metrics (e.g. CPU cores used).

  Returns:
    A dictionary holding parsed metrics.

  For example:
    metrics[key1][key2][measurement]

  metrics = {
  "width: 352": {
    "CPU cores used: 1.0": {
      "encode time (us)": [0.718005, 0.806925, 0.909726, 0.931835, 0.953642],
      "PSNR (dB)": [25.546029, 29.465518, 34.723535, 36.428493, 38.686551],
      "bitrate (kbps)": [50, 100, 300, 500, 1000]
    },
    "CPU cores used: 2.0": {
      "encode time (us)": [0.718005, 0.806925, 0.909726, 0.931835, 0.953642],
      "PSNR (dB)": [25.546029, 29.465518, 34.723535, 36.428493, 38.686551],
      "bitrate (kbps)": [50, 100, 300, 500, 1000]
    },
  },
  "width: 176": {
    "CPU cores used: 1.0": {
      "encode time (us)": [0.857897, 0.91608, 0.959173, 0.971116, 0.980961],
      "PSNR (dB)": [30.243646, 33.375592, 37.574387, 39.42184, 41.437897],
      "bitrate (kbps)": [50, 100, 300, 500, 1000]
    },
  }
  } """

    metrics = {}

    # Parse events.
    settings_file = open(filename)
    while True:
        line = settings_file.readline()
        if not line:
            break
        if re.search(r'%s' % EVENT_START, line):
            # Parse event.
            parsed = {}
            while True:
                line = settings_file.readline()
                if not line:
                    break
                if re.search(r'%s' % EVENT_END, line):
                    # Add parsed values to metrics.
                    key1 = setting1 + ': ' + str(parsed[setting1])
                    key2 = setting2 + ': ' + str(parsed[setting2])
                    if key1 not in metrics:
                        metrics[key1] = {}
                    if key2 not in metrics[key1]:
                        metrics[key1][key2] = {}

                    for label in parsed:
                        if label not in metrics[key1][key2]:
                            metrics[key1][key2][label] = []
                        metrics[key1][key2][label].append(parsed[label])

                    break

                TryFindMetric(parsed, line)

    settings_file.close()
    return metrics


def TryFindMetric(parsed, line):
    for metric in METRICS_TO_PARSE:
        name = metric[0]
        label = metric[1]
        if re.search(r'%s' % name, line):
            found, value = GetMetric(name, line)
            if found:
                parsed[label] = value
            return


def GetMetric(name, string):
    # Float (e.g. bitrate = 98.8253).
    pattern = r'%s\s*[:=]\s*([+-]?\d+\.*\d*)' % name
    m = re.search(r'%s' % pattern, string)
    if m is not None:
        return StringToFloat(m.group(1))

    # Alphanumeric characters (e.g. codec type : VP8).
    pattern = r'%s\s*[:=]\s*(\w+)' % name
    m = re.search(r'%s' % pattern, string)
    if m is not None:
        return True, m.group(1)

    return False, -1


def StringToFloat(value):
    try:
        value = float(value)
    except ValueError:
        print "Not a float, skipped %s" % value
        return False, -1

    return True, value


def Plot(y_metric, x_metric, metrics):
    """Plots y_metric vs x_metric per key in metrics.

  For example:
    y_metric = 'PSNR (dB)'
    x_metric = 'bitrate (kbps)'
    metrics = {
      "CPU cores used: 1.0": {
        "PSNR (dB)": [25.546029, 29.465518, 34.723535, 36.428493, 38.686551],
        "bitrate (kbps)": [50, 100, 300, 500, 1000]
      },
      "CPU cores used: 2.0": {
        "PSNR (dB)": [25.546029, 29.465518, 34.723535, 36.428493, 38.686551],
        "bitrate (kbps)": [50, 100, 300, 500, 1000]
      },
    }
    """
    for key in sorted(metrics):
        data = metrics[key]
        if y_metric not in data:
            print "Failed to find metric: %s" % y_metric
            continue

        y = numpy.array(data[y_metric])
        x = numpy.array(data[x_metric])
        if len(y) != len(x):
            print "Length mismatch for %s, %s" % (y, x)
            continue

        label = y_metric + ' - ' + str(key)

        plt.plot(x,
                 y,
                 label=label,
                 linewidth=1.5,
                 marker='o',
                 markersize=5,
                 markeredgewidth=0.0)


def PlotFigure(settings, y_metrics, x_metric, metrics, title):
    """Plots metrics in y_metrics list. One figure is plotted and each entry
  in the list is plotted in a subplot (and sorted per settings).

  For example:
    settings = ['width: 128.0', 'width: 160.0']. Sort subplot per setting.
    y_metrics = ['PSNR (dB)', 'PSNR (dB)']. Metric to plot per subplot.
    x_metric = 'bitrate (kbps)'

  """

    plt.figure()
    plt.suptitle(title, fontsize='large', fontweight='bold')
    settings.sort()
    rows = len(settings)
    cols = 1
    pos = 1
    while pos <= rows:
        plt.rc('grid', color=GRID_COLOR)
        ax = plt.subplot(rows, cols, pos)
        plt.grid()
        plt.setp(ax.get_xticklabels(), visible=(pos == rows), fontsize='large')
        plt.setp(ax.get_yticklabels(), fontsize='large')
        setting = settings[pos - 1]
        Plot(y_metrics[pos - 1], x_metric, metrics[setting])
        if setting.startswith(WIDTH[1]):
            plt.title(setting, fontsize='medium')
        plt.legend(fontsize='large', loc='best')
        pos += 1

    plt.xlabel(x_metric, fontsize='large')
    plt.subplots_adjust(left=0.06,
                        right=0.98,
                        bottom=0.05,
                        top=0.94,
                        hspace=0.08)


def GetTitle(filename, setting):
    title = ''
    if setting != CODEC_IMPLEMENTATION_NAME[1] and setting != CODEC_TYPE[1]:
        codec_types = ParseSetting(filename, CODEC_TYPE[1])
        for i in range(0, len(codec_types)):
            title += codec_types[i] + ', '

    if setting != CORES[1]:
        cores = ParseSetting(filename, CORES[1])
        for i in range(0, len(cores)):
            title += cores[i].split('.')[0] + ', '

    if setting != FRAMERATE[1]:
        framerate = ParseSetting(filename, FRAMERATE[1])
        for i in range(0, len(framerate)):
            title += framerate[i].split('.')[0] + ', '

    if (setting != CODEC_IMPLEMENTATION_NAME[1]
            and setting != ENCODER_IMPLEMENTATION_NAME[1]):
        enc_names = ParseSetting(filename, ENCODER_IMPLEMENTATION_NAME[1])
        for i in range(0, len(enc_names)):
            title += enc_names[i] + ', '

    if (setting != CODEC_IMPLEMENTATION_NAME[1]
            and setting != DECODER_IMPLEMENTATION_NAME[1]):
        dec_names = ParseSetting(filename, DECODER_IMPLEMENTATION_NAME[1])
        for i in range(0, len(dec_names)):
            title += dec_names[i] + ', '

    filenames = ParseSetting(filename, FILENAME[1])
    title += filenames[0].split('_')[0]

    num_frames = ParseSetting(filename, NUM_FRAMES[1])
    for i in range(0, len(num_frames)):
        title += ' (' + num_frames[i].split('.')[0] + ')'

    return title


def ToString(input_list):
    return ToStringWithoutMetric(input_list, ('', ''))


def ToStringWithoutMetric(input_list, metric):
    i = 1
    output_str = ""
    for m in input_list:
        if m != metric:
            output_str = output_str + ("%s. %s\n" % (i, m[1]))
            i += 1
    return output_str


def GetIdx(text_list):
    return int(raw_input(text_list)) - 1


def main():
    filename = sys.argv[1]

    # Setup.
    idx_metric = GetIdx("Choose metric:\n0. All\n%s" % ToString(RESULTS))
    if idx_metric == -1:
        # Plot all metrics. One subplot for each metric.
        # Per subplot: metric vs bitrate (per resolution).
        cores = ParseSetting(filename, CORES[1])
        setting1 = CORES[1]
        setting2 = WIDTH[1]
        sub_keys = [cores[0]] * len(Y_METRICS)
        y_metrics = Y_METRICS
        x_metric = BITRATE[1]
    else:
        resolutions = ParseSetting(filename, WIDTH[1])
        idx = GetIdx("Select metric for x-axis:\n%s" % ToString(X_SETTINGS))
        if X_SETTINGS[idx] == BITRATE:
            idx = GetIdx("Plot per:\n%s" %
                         ToStringWithoutMetric(SUBPLOT_SETTINGS, BITRATE))
            idx_setting = METRICS_TO_PARSE.index(SUBPLOT_SETTINGS[idx])
            # Plot one metric. One subplot for each resolution.
            # Per subplot: metric vs bitrate (per setting).
            setting1 = WIDTH[1]
            setting2 = METRICS_TO_PARSE[idx_setting][1]
            sub_keys = resolutions
            y_metrics = [RESULTS[idx_metric][1]] * len(sub_keys)
            x_metric = BITRATE[1]
        else:
            # Plot one metric. One subplot for each resolution.
            # Per subplot: metric vs setting (per bitrate).
            setting1 = WIDTH[1]
            setting2 = BITRATE[1]
            sub_keys = resolutions
            y_metrics = [RESULTS[idx_metric][1]] * len(sub_keys)
            x_metric = X_SETTINGS[idx][1]

    metrics = ParseMetrics(filename, setting1, setting2)

    # Stretch fig size.
    figsize = plt.rcParams["figure.figsize"]
    figsize[0] *= FIG_SIZE_SCALE_FACTOR_X
    figsize[1] *= FIG_SIZE_SCALE_FACTOR_Y
    plt.rcParams["figure.figsize"] = figsize

    PlotFigure(sub_keys, y_metrics, x_metric, metrics,
               GetTitle(filename, setting2))

    plt.show()


if __name__ == '__main__':
    main()



# Copyright (c) 2017 The WebRTC project authors. All Rights Reserved.
#
# Use of this source code is governed by a BSD-style license
# that can be found in the LICENSE file in the root of the source
# tree. An additional intellectual property rights grant can be found
# in the file PATENTS.  All contributing project authors may
# be found in the AUTHORS file in the root of the source tree.
"""TestDataGenerator factory class.
"""

import logging

from . import exceptions
from . import test_data_generation


class TestDataGeneratorFactory(object):
    """Factory class used to create test data generators.

  Usage: Create a factory passing parameters to the ctor with which the
  generators will be produced.
  """

    def __init__(self, aechen_ir_database_path, noise_tracks_path,
                 copy_with_identity):
        """Ctor.

    Args:
      aechen_ir_database_path: Path to the Aechen Impulse Response database.
      noise_tracks_path: Path to the noise tracks to add.
      copy_with_identity: Flag indicating whether the identity generator has to
                          make copies of the clean speech input files.
    """
        self._output_directory_prefix = None
        self._aechen_ir_database_path = aechen_ir_database_path
        self._noise_tracks_path = noise_tracks_path
        self._copy_with_identity = copy_with_identity

    def SetOutputDirectoryPrefix(self, prefix):
        self._output_directory_prefix = prefix

    def GetInstance(self, test_data_generators_class):
        """Creates an TestDataGenerator instance given a class object.

    Args:
      test_data_generators_class: TestDataGenerator class object (not an
                                  instance).

    Returns:
      TestDataGenerator instance.
    """
        if self._output_directory_prefix is None:
            raise exceptions.InitializationException(
                'The output directory prefix for test data generators is not set'
            )
        logging.debug('factory producing %s', test_data_generators_class)

        if test_data_generators_class == (
                test_data_generation.IdentityTestDataGenerator):
            return test_data_generation.IdentityTestDataGenerator(
                self._output_directory_prefix, self._copy_with_identity)
        elif test_data_generators_class == (
                test_data_generation.ReverberationTestDataGenerator):
            return test_data_generation.ReverberationTestDataGenerator(
                self._output_directory_prefix, self._aechen_ir_database_path)
        elif test_data_generators_class == (
                test_data_generation.AdditiveNoiseTestDataGenerator):
            return test_data_generation.AdditiveNoiseTestDataGenerator(
                self._output_directory_prefix, self._noise_tracks_path)
        else:
            return test_data_generators_class(self._output_directory_prefix)



# Copyright (c) 2017 The WebRTC project authors. All Rights Reserved.
#
# Use of this source code is governed by a BSD-style license
# that can be found in the LICENSE file in the root of the source
# tree. An additional intellectual property rights grant can be found
# in the file PATENTS.  All contributing project authors may
# be found in the AUTHORS file in the root of the source tree.
"""Unit tests for the annotations module.
"""

from __future__ import division
import logging
import os
import shutil
import tempfile
import unittest

import numpy as np

from . import annotations
from . import external_vad
from . import input_signal_creator
from . import signal_processing


class TestAnnotationsExtraction(unittest.TestCase):
    """Unit tests for the annotations module.
  """

    _CLEAN_TMP_OUTPUT = True
    _DEBUG_PLOT_VAD = False
    _VAD_TYPE_CLASS = annotations.AudioAnnotationsExtractor.VadType
    _ALL_VAD_TYPES = (_VAD_TYPE_CLASS.ENERGY_THRESHOLD
                      | _VAD_TYPE_CLASS.WEBRTC_COMMON_AUDIO
                      | _VAD_TYPE_CLASS.WEBRTC_APM)

    def setUp(self):
        """Create temporary folder."""
        self._tmp_path = tempfile.mkdtemp()
        self._wav_file_path = os.path.join(self._tmp_path, 'tone.wav')
        pure_tone, _ = input_signal_creator.InputSignalCreator.Create(
            'pure_tone', [440, 1000])
        signal_processing.SignalProcessingUtils.SaveWav(
            self._wav_file_path, pure_tone)
        self._sample_rate = pure_tone.frame_rate

    def tearDown(self):
        """Recursively delete temporary folder."""
        if self._CLEAN_TMP_OUTPUT:
            shutil.rmtree(self._tmp_path)
        else:
            logging.warning(self.id() + ' did not clean the temporary path ' +
                            (self._tmp_path))

    def testFrameSizes(self):
        e = annotations.AudioAnnotationsExtractor(self._ALL_VAD_TYPES)
        e.Extract(self._wav_file_path)
        samples_to_ms = lambda n, sr: 1000 * n // sr
        self.assertEqual(
            samples_to_ms(e.GetLevelFrameSize(), self._sample_rate),
            e.GetLevelFrameSizeMs())
        self.assertEqual(samples_to_ms(e.GetVadFrameSize(), self._sample_rate),
                         e.GetVadFrameSizeMs())

    def testVoiceActivityDetectors(self):
        for vad_type_value in range(0, self._ALL_VAD_TYPES + 1):
            vad_type = self._VAD_TYPE_CLASS(vad_type_value)
            e = annotations.AudioAnnotationsExtractor(vad_type=vad_type_value)
            e.Extract(self._wav_file_path)
            if vad_type.Contains(self._VAD_TYPE_CLASS.ENERGY_THRESHOLD):
                # pylint: disable=unpacking-non-sequence
                vad_output = e.GetVadOutput(
                    self._VAD_TYPE_CLASS.ENERGY_THRESHOLD)
                self.assertGreater(len(vad_output), 0)
                self.assertGreaterEqual(
                    float(np.sum(vad_output)) / len(vad_output), 0.95)

            if vad_type.Contains(self._VAD_TYPE_CLASS.WEBRTC_COMMON_AUDIO):
                # pylint: disable=unpacking-non-sequence
                vad_output = e.GetVadOutput(
                    self._VAD_TYPE_CLASS.WEBRTC_COMMON_AUDIO)
                self.assertGreater(len(vad_output), 0)
                self.assertGreaterEqual(
                    float(np.sum(vad_output)) / len(vad_output), 0.95)

            if vad_type.Contains(self._VAD_TYPE_CLASS.WEBRTC_APM):
                # pylint: disable=unpacking-non-sequence
                (vad_probs,
                 vad_rms) = e.GetVadOutput(self._VAD_TYPE_CLASS.WEBRTC_APM)
                self.assertGreater(len(vad_probs), 0)
                self.assertGreater(len(vad_rms), 0)
                self.assertGreaterEqual(
                    float(np.sum(vad_probs)) / len(vad_probs), 0.5)
                self.assertGreaterEqual(
                    float(np.sum(vad_rms)) / len(vad_rms), 20000)

            if self._DEBUG_PLOT_VAD:
                frame_times_s = lambda num_frames, frame_size_ms: np.arange(
                    num_frames).astype(np.float32) * frame_size_ms / 1000.0
                level = e.GetLevel()
                t_level = frame_times_s(num_frames=len(level),
                                        frame_size_ms=e.GetLevelFrameSizeMs())
                t_vad = frame_times_s(num_frames=len(vad_output),
                                      frame_size_ms=e.GetVadFrameSizeMs())
                import matplotlib.pyplot as plt
                plt.figure()
                plt.hold(True)
                plt.plot(t_level, level)
                plt.plot(t_vad, vad_output * np.max(level), '.')
                plt.show()

    def testSaveLoad(self):
        e = annotations.AudioAnnotationsExtractor(self._ALL_VAD_TYPES)
        e.Extract(self._wav_file_path)
        e.Save(self._tmp_path, "fake-annotation")

        data = np.load(
            os.path.join(
                self._tmp_path,
                e.GetOutputFileNameTemplate().format("fake-annotation")))
        np.testing.assert_array_equal(e.GetLevel(), data['level'])
        self.assertEqual(np.float32, data['level'].dtype)
        np.testing.assert_array_equal(
            e.GetVadOutput(self._VAD_TYPE_CLASS.ENERGY_THRESHOLD),
            data['vad_energy_output'])
        np.testing.assert_array_equal(
            e.GetVadOutput(self._VAD_TYPE_CLASS.WEBRTC_COMMON_AUDIO),
            data['vad_output'])
        np.testing.assert_array_equal(
            e.GetVadOutput(self._VAD_TYPE_CLASS.WEBRTC_APM)[0],
            data['vad_probs'])
        np.testing.assert_array_equal(
            e.GetVadOutput(self._VAD_TYPE_CLASS.WEBRTC_APM)[1],
            data['vad_rms'])
        self.assertEqual(np.uint8, data['vad_energy_output'].dtype)
        self.assertEqual(np.float64, data['vad_probs'].dtype)
        self.assertEqual(np.float64, data['vad_rms'].dtype)

    def testEmptyExternalShouldNotCrash(self):
        for vad_type_value in range(0, self._ALL_VAD_TYPES + 1):
            annotations.AudioAnnotationsExtractor(vad_type_value, {})

    def testFakeExternalSaveLoad(self):
        def FakeExternalFactory():
            return external_vad.ExternalVad(
                os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'fake_external_vad.py'), 'fake')

        for vad_type_value in range(0, self._ALL_VAD_TYPES + 1):
            e = annotations.AudioAnnotationsExtractor(
                vad_type_value, {'fake': FakeExternalFactory()})
            e.Extract(self._wav_file_path)
            e.Save(self._tmp_path, annotation_name="fake-annotation")
            data = np.load(
                os.path.join(
                    self._tmp_path,
                    e.GetOutputFileNameTemplate().format("fake-annotation")))
            self.assertEqual(np.float32, data['extvad_conf-fake'].dtype)
            np.testing.assert_almost_equal(np.arange(100, dtype=np.float32),
                                           data['extvad_conf-fake'])



#!/usr/bin/env python
# Copyright 2019 The Chromium Authors. All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.

from __future__ import division
from __future__ import print_function

from array import array
import os
import random
import shutil
import sys

MAX_INPUT_SIZE = 5000  # < 1 MB so we don't blow up fuzzer build sizes.
MAX_FLOAT32 = 3.4028235e+38


def IsValidSize(n):
  if n == 0:
    return False
  # PFFFT only supports transforms for inputs of length N of the form
  # N = (2^a)*(3^b)*(5^c) where a >= 5, b >=0, c >= 0.
  FACTORS = [2, 3, 5]
  factorization = [0, 0, 0]
  for i, factor in enumerate(FACTORS):
    while n % factor == 0:
      n = n // factor
      factorization[i] += 1
  return factorization[0] >= 5 and n == 1


def WriteFloat32ArrayToFile(file_path, size, generator):
  """Generate an array of float32 values and writes to file."""
  with open(file_path, 'wb') as f:
    float_array = array('f', [generator() for _ in range(size)])
    float_array.tofile(f)


def main():
  if len(sys.argv) < 2:
    print('Usage: %s <path to output directory>' % sys.argv[0])
    sys.exit(1)

  output_path = sys.argv[1]
  # Start with a clean output directory.
  if os.path.exists(output_path):
    shutil.rmtree(output_path)
  os.makedirs(output_path)

  # List of valid input sizes.
  N = [n for n in range(MAX_INPUT_SIZE) if IsValidSize(n)]

  # Set the seed to always generate the same random data.
  random.seed(0)

  # Generate different types of input arrays for each target length.
  for n in N:
    # Zeros.
    WriteFloat32ArrayToFile(
        os.path.join(output_path, 'zeros_%d' % n), n, lambda: 0)
    # Max float 32.
    WriteFloat32ArrayToFile(
        os.path.join(output_path, 'max_%d' % n), n, lambda: MAX_FLOAT32)
    # Random values in the s16 range.
    rnd_s16 = lambda: 32768.0 * 2.0 * (random.random() - 0.5)
    WriteFloat32ArrayToFile(
        os.path.join(output_path, 'rnd_s16_%d' % n), n, rnd_s16)

  sys.exit(0)


if __name__ == '__main__':
  main()



# Copyright (c) 2017 The WebRTC project authors. All Rights Reserved.
#
# Use of this source code is governed by a BSD-style license
# that can be found in the LICENSE file in the root of the source
# tree. An additional intellectual property rights grant can be found
# in the file PATENTS.  All contributing project authors may
# be found in the AUTHORS file in the root of the source tree.
"""Imports a filtered subset of the scores and configurations computed
by apm_quality_assessment.py into a pandas data frame.
"""

import argparse
import glob
import logging
import os
import re
import sys

try:
    import pandas as pd
except ImportError:
    logging.critical('Cannot import the third-party Python package pandas')
    sys.exit(1)

from . import data_access as data_access
from . import simulation as sim

# Compiled regular expressions used to extract score descriptors.
RE_CONFIG_NAME = re.compile(sim.ApmModuleSimulator.GetPrefixApmConfig() +
                            r'(.+)')
RE_CAPTURE_NAME = re.compile(sim.ApmModuleSimulator.GetPrefixCapture() +
                             r'(.+)')
RE_RENDER_NAME = re.compile(sim.ApmModuleSimulator.GetPrefixRender() + r'(.+)')
RE_ECHO_SIM_NAME = re.compile(sim.ApmModuleSimulator.GetPrefixEchoSimulator() +
                              r'(.+)')
RE_TEST_DATA_GEN_NAME = re.compile(
    sim.ApmModuleSimulator.GetPrefixTestDataGenerator() + r'(.+)')
RE_TEST_DATA_GEN_PARAMS = re.compile(
    sim.ApmModuleSimulator.GetPrefixTestDataGeneratorParameters() + r'(.+)')
RE_SCORE_NAME = re.compile(sim.ApmModuleSimulator.GetPrefixScore() +
                           r'(.+)(\..+)')


def InstanceArgumentsParser():
    """Arguments parser factory.
  """
    parser = argparse.ArgumentParser(
        description=('Override this description in a user script by changing'
                     ' `parser.description` of the returned parser.'))

    parser.add_argument('-o',
                        '--output_dir',
                        required=True,
                        help=('the same base path used with the '
                              'apm_quality_assessment tool'))

    parser.add_argument(
        '-c',
        '--config_names',
        type=re.compile,
        help=('regular expression to filter the APM configuration'
              ' names'))

    parser.add_argument(
        '-i',
        '--capture_names',
        type=re.compile,
        help=('regular expression to filter the capture signal '
              'names'))

    parser.add_argument('-r',
                        '--render_names',
                        type=re.compile,
                        help=('regular expression to filter the render signal '
                              'names'))

    parser.add_argument(
        '-e',
        '--echo_simulator_names',
        type=re.compile,
        help=('regular expression to filter the echo simulator '
              'names'))

    parser.add_argument('-t',
                        '--test_data_generators',
                        type=re.compile,
                        help=('regular expression to filter the test data '
                              'generator names'))

    parser.add_argument(
        '-s',
        '--eval_scores',
        type=re.compile,
        help=('regular expression to filter the evaluation score '
              'names'))

    return parser


def _GetScoreDescriptors(score_filepath):
    """Extracts a score descriptor from the given score file path.

  Args:
    score_filepath: path to the score file.

  Returns:
    A tuple of strings (APM configuration name, capture audio track name,
    render audio track name, echo simulator name, test data generator name,
    test data generator parameters as string, evaluation score name).
  """
    fields = score_filepath.split(os.sep)[-7:]
    extract_name = lambda index, reg_expr: (reg_expr.match(fields[index]).
                                            groups(0)[0])
    return (
        extract_name(0, RE_CONFIG_NAME),
        extract_name(1, RE_CAPTURE_NAME),
        extract_name(2, RE_RENDER_NAME),
        extract_name(3, RE_ECHO_SIM_NAME),
        extract_name(4, RE_TEST_DATA_GEN_NAME),
        extract_name(5, RE_TEST_DATA_GEN_PARAMS),
        extract_name(6, RE_SCORE_NAME),
    )


def _ExcludeScore(config_name, capture_name, render_name, echo_simulator_name,
                  test_data_gen_name, score_name, args):
    """Decides whether excluding a score.

  A set of optional regular expressions in args is used to determine if the
  score should be excluded (depending on its |*_name| descriptors).

  Args:
    config_name: APM configuration name.
    capture_name: capture audio track name.
    render_name: render audio track name.
    echo_simulator_name: echo simulator name.
    test_data_gen_name: test data generator name.
    score_name: evaluation score name.
    args: parsed arguments.

  Returns:
    A boolean.
  """
    value_regexpr_pairs = [
        (config_name, args.config_names),
        (capture_name, args.capture_names),
        (render_name, args.render_names),
        (echo_simulator_name, args.echo_simulator_names),
        (test_data_gen_name, args.test_data_generators),
        (score_name, args.eval_scores),
    ]

    # Score accepted if each value matches the corresponding regular expression.
    for value, regexpr in value_regexpr_pairs:
        if regexpr is None:
            continue
        if not regexpr.match(value):
            return True

    return False


def FindScores(src_path, args):
    """Given a search path, find scores and return a DataFrame object.

  Args:
    src_path: Search path pattern.
    args: parsed arguments.

  Returns:
    A DataFrame object.
  """
    # Get scores.
    scores = []
    for score_filepath in glob.iglob(src_path):
        # Extract score descriptor fields from the path.
        (config_name, capture_name, render_name, echo_simulator_name,
         test_data_gen_name, test_data_gen_params,
         score_name) = _GetScoreDescriptors(score_filepath)

        # Ignore the score if required.
        if _ExcludeScore(config_name, capture_name, render_name,
                         echo_simulator_name, test_data_gen_name, score_name,
                         args):
            logging.info('ignored score: %s %s %s %s %s %s', config_name,
                         capture_name, render_name, echo_simulator_name,
                         test_data_gen_name, score_name)
            continue

        # Read metadata and score.
        metadata = data_access.Metadata.LoadAudioTestDataPaths(
            os.path.split(score_filepath)[0])
        score = data_access.ScoreFile.Load(score_filepath)

        # Add a score with its descriptor fields.
        scores.append((
            metadata['clean_capture_input_filepath'],
            metadata['echo_free_capture_filepath'],
            metadata['echo_filepath'],
            metadata['render_filepath'],
            metadata['capture_filepath'],
            metadata['apm_output_filepath'],
            metadata['apm_reference_filepath'],
            config_name,
            capture_name,
            render_name,
            echo_simulator_name,
            test_data_gen_name,
            test_data_gen_params,
            score_name,
            score,
        ))

    return pd.DataFrame(data=scores,
                        columns=(
                            'clean_capture_input_filepath',
                            'echo_free_capture_filepath',
                            'echo_filepath',
                            'render_filepath',
                            'capture_filepath',
                            'apm_output_filepath',
                            'apm_reference_filepath',
                            'apm_config',
                            'capture',
                            'render',
                            'echo_simulator',
                            'test_data_gen',
                            'test_data_gen_params',
                            'eval_score_name',
                            'score',
                        ))


def ConstructSrcPath(args):
    return os.path.join(
        args.output_dir,
        sim.ApmModuleSimulator.GetPrefixApmConfig() + '*',
        sim.ApmModuleSimulator.GetPrefixCapture() + '*',
        sim.ApmModuleSimulator.GetPrefixRender() + '*',
        sim.ApmModuleSimulator.GetPrefixEchoSimulator() + '*',
        sim.ApmModuleSimulator.GetPrefixTestDataGenerator() + '*',
        sim.ApmModuleSimulator.GetPrefixTestDataGeneratorParameters() + '*',
        sim.ApmModuleSimulator.GetPrefixScore() + '*')



#!/usr/bin/env python3

# Copyright 2016 The LibYuv Project Authors. All rights reserved.
#
# Use of this source code is governed by a BSD-style license
# that can be found in the LICENSE file in the root of the source
# tree. An additional intellectual property rights grant can be found
# in the file PATENTS. All contributing project authors may
# be found in the AUTHORS file in the root of the source tree.

"""
This file emits the list of reasons why a particular build needs to be clobbered
(or a list of 'landmines').
"""

import sys


def print_landmines():
  """
  ALL LANDMINES ARE EMITTED FROM HERE.
  """
  # DO NOT add landmines as part of a regular CL. Landmines are a last-effort
  # bandaid fix if a CL that got landed has a build dependency bug and all bots
  # need to be cleaned up. If you're writing a new CL that causes build
  # dependency problems, fix the dependency problems instead of adding a
  # landmine.
  # See the Chromium version in src/build/get_landmines.py for usage examples.
  print('Clobber to remove GYP artifacts after switching bots to GN.')
  print('Another try to remove GYP artifacts after switching bots to GN.')


def main():
  print_landmines()
  return 0


if __name__ == '__main__':
  sys.exit(main())



#!/usr/bin/env vpython3
# Copyright (c) 2017 The WebRTC project authors. All Rights Reserved.
#
# Use of this source code is governed by a BSD-style license
# that can be found in the LICENSE file in the root of the source
# tree. An additional intellectual property rights grant can be found
# in the file PATENTS.  All contributing project authors may
# be found in the AUTHORS file in the root of the source tree.
"""
This script is the wrapper that runs the low-bandwidth audio test.

After running the test, post-process steps for calculating audio quality of the
output files will be performed.
"""

import argparse
import collections
import json
import logging
import os
import re
import shutil
import subprocess
import sys

SCRIPT_DIR = os.path.dirname(os.path.abspath(__file__))
SRC_DIR = os.path.normpath(os.path.join(SCRIPT_DIR, os.pardir, os.pardir))

NO_TOOLS_ERROR_MESSAGE = (
    'Could not find PESQ or POLQA at %s.\n'
    '\n'
    'To fix this run:\n'
    '  python %s %s\n'
    '\n'
    'Note that these tools are Google-internal due to licensing, so in order '
    'to use them you will have to get your own license and manually put them '
    'in the right location.\n'
    'See https://cs.chromium.org/chromium/src/third_party/webrtc/tools_webrtc/'
    'download_tools.py?rcl=bbceb76f540159e2dba0701ac03c514f01624130&l=13')


def _LogCommand(command):
  logging.info('Running %r', command)
  return command


def _ParseArgs():
  parser = argparse.ArgumentParser(description='Run low-bandwidth audio tests.')
  parser.add_argument('build_dir',
                      help='Path to the build directory (e.g. out/Release).')
  parser.add_argument('--remove',
                      action='store_true',
                      help='Remove output audio files after testing.')
  parser.add_argument(
      '--android',
      action='store_true',
      help='Perform the test on a connected Android device instead.')
  parser.add_argument('--adb-path', help='Path to adb binary.', default='adb')
  parser.add_argument('--num-retries',
                      default='0',
                      help='Number of times to retry the test on Android.')
  parser.add_argument(
      '--isolated-script-test-perf-output',
      default=None,
      help='Path to store perf results in histogram proto format.')
  parser.add_argument(
      '--isolated-script-test-output',
      default=None,
      help='Path to output an empty JSON file which Chromium infra requires.')

  return parser.parse_known_args()


def _GetPlatform():
  if sys.platform == 'win32':
    return 'win'
  if sys.platform == 'darwin':
    return 'mac'
  if sys.platform.startswith('linux'):
    return 'linux'
  raise AssertionError('Unknown platform %s' % sys.platform)


def _GetExtension():
  return '.exe' if sys.platform == 'win32' else ''


def _GetPathToTools():
  tools_dir = os.path.join(SRC_DIR, 'tools_webrtc')
  toolchain_dir = os.path.join(tools_dir, 'audio_quality')

  platform = _GetPlatform()
  ext = _GetExtension()

  pesq_path = os.path.join(toolchain_dir, platform, 'pesq' + ext)
  if not os.path.isfile(pesq_path):
    pesq_path = None

  polqa_path = os.path.join(toolchain_dir, platform, 'PolqaOem64' + ext)
  if not os.path.isfile(polqa_path):
    polqa_path = None

  if (platform != 'mac' and not polqa_path) or not pesq_path:
    logging.error(NO_TOOLS_ERROR_MESSAGE, toolchain_dir,
                  os.path.join(tools_dir, 'download_tools.py'), toolchain_dir)

  return pesq_path, polqa_path


def ExtractTestRuns(lines, echo=False):
  """Extracts information about tests from the output of a test runner.

  Produces tuples
  (android_device, test_name, reference_file, degraded_file, cur_perf_results).
  """
  for line in lines:
    if echo:
      sys.stdout.write(line)

    # Output from Android has a prefix with the device name.
    android_prefix_re = r'(?:I\b.+\brun_tests_on_device\((.+?)\)\s*)?'
    test_re = r'^' + android_prefix_re + (r'TEST (\w+) ([^ ]+?) ([^\s]+)'
                                          r' ?([^\s]+)?\s*$')

    match = re.search(test_re, line)
    if match:
      yield match.groups()


def _GetFile(file_path,
             out_dir,
             move=False,
             android=False,
             adb_prefix=('adb', )):
  out_file_name = os.path.basename(file_path)
  out_file_path = os.path.join(out_dir, out_file_name)

  if android:
    # Pull the file from the connected Android device.
    adb_command = adb_prefix + ('pull', file_path, out_dir)
    subprocess.check_call(_LogCommand(adb_command))
    if move:
      # Remove that file.
      adb_command = adb_prefix + ('shell', 'rm', file_path)
      subprocess.check_call(_LogCommand(adb_command))
  elif os.path.abspath(file_path) != os.path.abspath(out_file_path):
    if move:
      shutil.move(file_path, out_file_path)
    else:
      shutil.copy(file_path, out_file_path)

  return out_file_path


def _RunPesq(executable_path,
             reference_file,
             degraded_file,
             sample_rate_hz=16000):
  directory = os.path.dirname(reference_file)
  assert os.path.dirname(degraded_file) == directory

  # Analyze audio.
  command = [
      executable_path,
      '+%d' % sample_rate_hz,
      os.path.basename(reference_file),
      os.path.basename(degraded_file)
  ]
  # Need to provide paths in the current directory due to a bug in PESQ:
  # On Mac, for some 'path/to/file.wav', if 'file.wav' is longer than
  # 'path/to', PESQ crashes.
  out = subprocess.check_output(_LogCommand(command),
                                cwd=directory,
                                universal_newlines=True,
                                stderr=subprocess.STDOUT)

  # Find the scores in stdout of PESQ.
  match = re.search(
      r'Prediction \(Raw MOS, MOS-LQO\):\s+=\s+([\d.]+)\s+([\d.]+)', out)
  if match:
    raw_mos, _ = match.groups()
    return {'pesq_mos': (raw_mos, 'unitless')}
  logging.error('PESQ: %s', out.splitlines()[-1])
  return {}


def _RunPolqa(executable_path, reference_file, degraded_file):
  # Analyze audio.
  command = [
      executable_path, '-q', '-LC', 'NB', '-Ref', reference_file, '-Test',
      degraded_file
  ]
  process = subprocess.Popen(_LogCommand(command),
                             universal_newlines=True,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)
  out, err = process.communicate()

  # Find the scores in stdout of POLQA.
  match = re.search(r'\bMOS-LQO:\s+([\d.]+)', out)

  if process.returncode != 0 or not match:
    if process.returncode == 2:
      logging.warning('%s (2)', err.strip())
      logging.warning('POLQA license error, skipping test.')
    else:
      logging.error('%s (%d)', err.strip(), process.returncode)
    return {}

  mos_lqo, = match.groups()
  return {'polqa_mos_lqo': (mos_lqo, 'unitless')}


def _MergeInPerfResultsFromCcTests(histograms, run_perf_results_file):
  from tracing.value import histogram_set

  cc_histograms = histogram_set.HistogramSet()
  with open(run_perf_results_file, 'rb') as f:
    contents = f.read()
    if not contents:
      return

    cc_histograms.ImportProto(contents)

  histograms.Merge(cc_histograms)


Analyzer = collections.namedtuple(
    'Analyzer', ['name', 'func', 'executable', 'sample_rate_hz'])


def _ConfigurePythonPath(args):
  script_dir = os.path.dirname(os.path.realpath(__file__))
  checkout_root = os.path.abspath(os.path.join(script_dir, os.pardir,
                                               os.pardir))

  # TODO(https://crbug.com/1029452): Use a copy rule and add these from the
  # out dir like for the third_party/protobuf code.
  sys.path.insert(
      0, os.path.join(checkout_root, 'third_party', 'catapult', 'tracing'))

  # The low_bandwidth_audio_perf_test gn rule will build the protobuf stub
  # for python, so put it in the path for this script before we attempt to
  # import it.
  histogram_proto_path = os.path.join(os.path.abspath(args.build_dir),
                                      'pyproto', 'tracing', 'tracing', 'proto')
  sys.path.insert(0, histogram_proto_path)
  proto_stub_path = os.path.join(os.path.abspath(args.build_dir), 'pyproto')
  sys.path.insert(0, proto_stub_path)

  # Fail early in case the proto hasn't been built.
  try:
    #pylint: disable=unused-import
    import histogram_pb2
  except ImportError as e:
    raise ImportError('Could not import histogram_pb2. You need to build the '
                      'low_bandwidth_audio_perf_test target before invoking '
                      'this script. Expected to find '
                      'histogram_pb2.py in %s.' % histogram_proto_path) from e


def main():
  logging.basicConfig(format='%(asctime)s %(levelname)-8s %(message)s',
                      level=logging.INFO,
                      datefmt='%Y-%m-%d %H:%M:%S')
  logging.info('Invoked with %s', str(sys.argv))

  args, extra_test_args = _ParseArgs()

  _ConfigurePythonPath(args)

  # Import catapult modules here after configuring the pythonpath.
  from tracing.value import histogram_set
  from tracing.value.diagnostics import reserved_infos
  from tracing.value.diagnostics import generic_set

  pesq_path, polqa_path = _GetPathToTools()
  if pesq_path is None:
    return 1

  out_dir = os.path.join(args.build_dir, '..')
  if args.android:
    test_command = [
        os.path.join(args.build_dir, 'bin', 'run_low_bandwidth_audio_test'),
        '-v', '--num-retries', args.num_retries
    ]
  else:
    test_command = [os.path.join(args.build_dir, 'low_bandwidth_audio_test')]

  analyzers = [Analyzer('pesq', _RunPesq, pesq_path, 16000)]
  # Check if POLQA can run at all, or skip the 48 kHz tests entirely.
  example_path = os.path.join(SRC_DIR, 'resources', 'voice_engine',
                              'audio_tiny48.wav')
  if polqa_path and _RunPolqa(polqa_path, example_path, example_path):
    analyzers.append(Analyzer('polqa', _RunPolqa, polqa_path, 48000))

  histograms = histogram_set.HistogramSet()
  for analyzer in analyzers:
    # Start the test executable that produces audio files.
    test_process = subprocess.Popen(_LogCommand(test_command + [
        '--sample_rate_hz=%d' % analyzer.sample_rate_hz,
        '--test_case_prefix=%s' % analyzer.name,
    ] + extra_test_args),
                                    universal_newlines=True,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.STDOUT)
    perf_results_file = None
    try:
      lines = iter(test_process.stdout.readline, '')
      for result in ExtractTestRuns(lines, echo=True):
        (android_device, test_name, reference_file, degraded_file,
         perf_results_file) = result

        adb_prefix = (args.adb_path, )
        if android_device:
          adb_prefix += ('-s', android_device)

        reference_file = _GetFile(reference_file,
                                  out_dir,
                                  android=args.android,
                                  adb_prefix=adb_prefix)
        degraded_file = _GetFile(degraded_file,
                                 out_dir,
                                 move=True,
                                 android=args.android,
                                 adb_prefix=adb_prefix)

        analyzer_results = analyzer.func(analyzer.executable, reference_file,
                                         degraded_file)
        for metric, (value, units) in list(analyzer_results.items()):
          hist = histograms.CreateHistogram(metric, units, [value])
          user_story = generic_set.GenericSet([test_name])
          hist.diagnostics[reserved_infos.STORIES.name] = user_story

          # Output human readable results.
          print('RESULT %s: %s= %s %s' % (metric, test_name, value, units))

        if args.remove:
          os.remove(reference_file)
          os.remove(degraded_file)
    finally:
      test_process.terminate()
    if perf_results_file:
      perf_results_file = _GetFile(perf_results_file,
                                   out_dir,
                                   move=True,
                                   android=args.android,
                                   adb_prefix=adb_prefix)
      _MergeInPerfResultsFromCcTests(histograms, perf_results_file)
      if args.remove:
        os.remove(perf_results_file)

  if args.isolated_script_test_perf_output:
    with open(args.isolated_script_test_perf_output, 'wb') as f:
      f.write(histograms.AsProto().SerializeToString())

  if args.isolated_script_test_output:
    with open(args.isolated_script_test_output, 'w') as f:
      json.dump({"version": 3}, f)

  return test_process.wait()


if __name__ == '__main__':
  sys.exit(main())



#!/usr/bin/env python
# Copyright (c) 2012 The Chromium Authors. All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.

"""Creates Windows and POSIX stub files for a given set of signatures.

For libraries that need to be loaded outside of the standard executable startup
path mechanism, stub files need to be generated for the wanted functions.  In
Windows, this is done via "def" files and the delay load mechanism.  On a POSIX
system, a set of stub functions need to be generated that dispatch to functions
found via dlsym.

This script takes a set of files, where each file is a list of C-style
signatures (one signature per line).  The output is either a Windows def file,
or a header + implementation file of stubs suitable for use in a POSIX system.

This script also handles variadic functions, e.g.
void printf(const char* s, ...);

TODO(hclam): Fix the situation for variadic functions.
Stub for the above function will be generated and inside the stub function it
is translated to:
void printf(const char* s, ...) {
  printf_ptr(s, (void*)arg1);
}

Only one argument from the variadic arguments is used and it will be used as
type void*.
"""

__author__ = 'ajwong@chromium.org (Albert J. Wong)'

import optparse
import os
import re
import string
import subprocess
import sys


class Error(Exception):
  pass


class BadSignatureError(Error):
  pass


class SubprocessError(Error):
  def __init__(self, message, error_code):
    Error.__init__(self)
    self.message = message
    self.error_code = error_code

  def __str__(self):
    return 'Failed with code %s: %s' % (self.message, repr(self.error_code))


# Regular expression used to parse function signatures in the input files.
# The regex is built around identifying the "identifier" for the function name.
# We consider the identifier to be the string that follows these constraints:
#
#   1) Starts with [_a-ZA-Z] (C++ spec 2.10).
#   2) Continues with [_a-ZA-Z0-9] (C++ spec 2.10).
#   3) Precedes an opening parenthesis by 0 or more whitespace chars.
#
# From that, all preceding characters are considered the return value.
# Trailing characters should have a substring matching the form (.*).  That
# is considered the arguments.
SIGNATURE_REGEX = re.compile('(?P<return_type>.+?)'
                             '(?P<name>[_a-zA-Z][_a-zA-Z0-9]+)\s*'
                             '\((?P<params>.*?)\)')

# Used for generating C++ identifiers.
INVALID_C_IDENT_CHARS = re.compile('[^_a-zA-Z0-9]')

# Constants defining the supported file types options.
FILE_TYPE_WIN_X86 = 'windows_lib'
FILE_TYPE_WIN_X64 = 'windows_lib_x64'
FILE_TYPE_POSIX_STUB = 'posix_stubs'
FILE_TYPE_WIN_DEF = 'windows_def'

# Template for generating a stub function definition.  Includes a forward
# declaration marking the symbol as weak.  This template takes the following
# named parameters.
#   return_type: The return type.
#   export: The macro used to alter the stub's visibility.
#   name: The name of the function.
#   params: The parameters to the function.
#   return_prefix: 'return ' if this function is not void. '' otherwise.
#   arg_list: The arguments used to call the stub function.
STUB_FUNCTION_DEFINITION = (
    """extern %(return_type)s %(name)s(%(params)s) __attribute__((weak));
DISABLE_CFI_ICALL
%(return_type)s %(export)s %(name)s(%(params)s) {
  %(return_prefix)s%(name)s_ptr(%(arg_list)s);
}""")

# Template for generating a variadic stub function definition with return
# value.
# Includes a forward declaration marking the symbol as weak.
# This template takes the following named parameters.
#   return_type: The return type.
#   export: The macro used to alter the stub's visibility.
#   name: The name of the function.
#   params: The parameters to the function.
#   arg_list: The arguments used to call the stub function without the
#             variadic argument.
#   last_named_arg: Name of the last named argument before the variadic
#                   argument.
VARIADIC_STUB_FUNCTION_DEFINITION = (
    """extern %(return_type)s %(name)s(%(params)s) __attribute__((weak));
DISABLE_CFI_ICALL
%(return_type)s %(export)s %(name)s(%(params)s) {
  va_list args___;
  va_start(args___, %(last_named_arg)s);
  %(return_type)s ret___ = %(name)s_ptr(%(arg_list)s, va_arg(args___, void*));
  va_end(args___);
  return ret___;
}""")

# Template for generating a variadic stub function definition without
# return value.
# Includes a forward declaration marking the symbol as weak.
# This template takes the following named parameters.
#   name: The name of the function.
#   params: The parameters to the function.
#   export: The macro used to alter the stub's visibility.
#   arg_list: The arguments used to call the stub function without the
#             variadic argument.
#   last_named_arg: Name of the last named argument before the variadic
#                   argument.
VOID_VARIADIC_STUB_FUNCTION_DEFINITION = (
    """extern void %(name)s(%(params)s) __attribute__((weak));
DISABLE_CFI_ICALL
void %(export)s %(name)s(%(params)s) {
  va_list args___;
  va_start(args___, %(last_named_arg)s);
  %(name)s_ptr(%(arg_list)s, va_arg(args___, void*));
  va_end(args___);
}""")

# Template for the preamble for the stub header file with the header guards,
# standard set of includes, and namespace opener.  This template takes the
# following named parameters:
#   guard_name: The macro to use as the header guard.
#   namespace: The namespace for the stub functions.
#   logging_include: Header file where the logging function is defined.
STUB_HEADER_PREAMBLE = """// This is generated file. Do not modify directly.

#ifndef %(guard_name)s
#define %(guard_name)s

#include <stdarg.h>
#include <map>
#include <string>
#include <vector>

#include "%(logging_include)s"

namespace %(namespace)s {
"""

# Template for the end of the stub header. This closes the namespace and the
# header guards.  This template takes the following named parameters:
#   guard_name: The macro to use as the header guard.
#   namespace: The namespace for the stub functions.
STUB_HEADER_CLOSER = """}  // namespace %(namespace)s

#endif  // %(guard_name)s
"""

# The standard includes needed for the stub implementation file.  Takes one
# string substitution with the path to the associated stub header file.
IMPLEMENTATION_PREAMBLE = """// This is generated file. Do not modify directly.

#include "%s"

#include <stdlib.h>  // For NULL.
#include <dlfcn.h>   // For dlsym, dlopen.

#include <map>
#include <vector>

#include "base/compiler_specific.h"
"""

# The start and end templates for the enum definitions used by the Umbrella
# initializer.
UMBRELLA_ENUM_START = """// Enum and typedef for umbrella initializer.
enum StubModules {
"""
UMBRELLA_ENUM_END = """  kNumStubModules
};

"""

# Start and end of the extern "C" section for the implementation contents.
IMPLEMENTATION_CONTENTS_C_START = """extern "C" {

"""
IMPLEMENTATION_CONTENTS_C_END = """
}  // extern "C"


"""

# Templates for the start and end of a namespace.  Takes one parameter, the
# namespace name.
NAMESPACE_START = """namespace %s {

"""
NAMESPACE_END = """}  // namespace %s

"""

# Comment to include before the section declaring all the function pointers
# used by the stub functions.
FUNCTION_POINTER_SECTION_COMMENT = (
    """// Static pointers that will hold the location of the real function
// implementations after the module has been loaded.
""")

# Template for the module initialization check function.  This template
# takes two parameters: the function name, and the conditional used to
# verify the module's initialization.
MODULE_INITIALIZATION_CHECK_FUNCTION = (
    """// Returns true if all stubs have been properly initialized.
bool %s() {
  if (%s) {
    return true;
  } else {
    return false;
  }
}

""")

# Template for the line that initialize the stub pointer.  This template takes
# the following named parameters:
#   name: The name of the function.
#   return_type: The return type.
#   params: The parameters to the function.
#   logging_function: Function call for error logging.
STUB_POINTER_INITIALIZER = """  %(name)s_ptr =
    reinterpret_cast<%(return_type)s (*)(%(parameters)s)>(
      dlsym(module, "%(name)s"));
  if (!%(name)s_ptr) {
    %(logging_function)s << "Couldn't load %(name)s, dlerror() says:\\n"
      << dlerror();
  }
"""

# Template for module initializer function start and end.  This template takes
# one parameter which is the initializer function name.
MODULE_INITIALIZE_START = """// Initializes the module stubs.
void %s(void* module) {
"""
MODULE_INITIALIZE_END = """}

"""

# Template for module uninitializer function start and end.  This template
# takes one parameter which is the initializer function name.
MODULE_UNINITIALIZE_START = (
    """// Uninitialize the module stubs.  Reset pointers to NULL.
void %s() {
""")
MODULE_UNINITIALIZE_END = """}

"""


# Open namespace and add typedef for internal data structures used by the
# umbrella initializer.
UMBRELLA_INITIALIZER_START = """namespace %s {
typedef std::map<StubModules, void*> StubHandleMap;
"""

# Function close DSOs on error and clean up dangling references.
UMBRELLA_INITIALIZER_CLEANUP_FUNCTION = (
    """static void CloseLibraries(StubHandleMap* stub_handles) {
  for (StubHandleMap::const_iterator it = stub_handles->begin();
       it != stub_handles->end();
       ++it) {
    dlclose(it->second);
  }

  stub_handles->clear();
}
""")

# Function to initialize each DSO for the given paths.
#   logging_function: Function call that will be used for error logging.
UMBRELLA_INITIALIZER_INITIALIZE_FUNCTION_START = (
    """bool InitializeStubs(const StubPathMap& path_map) {
  StubHandleMap opened_libraries;
  for (int i = 0; i < kNumStubModules; ++i) {
    StubModules cur_module = static_cast<StubModules>(i);
    // If a module is missing, we fail.
    StubPathMap::const_iterator it = path_map.find(cur_module);
    if (it == path_map.end()) {
      CloseLibraries(&opened_libraries);
      return false;
    }

    // Otherwise, attempt to dlopen the library.
    const std::vector<std::string>& paths = it->second;
    bool module_opened = false;
    for (std::vector<std::string>::const_iterator dso_path = paths.begin();
         !module_opened && dso_path != paths.end();
         ++dso_path) {
      void* handle = dlopen(dso_path->c_str(), RTLD_LAZY);
      if (handle != NULL) {
        module_opened = true;
        opened_libraries[cur_module] = handle;
      } else {
        %(logging_function)s << "dlopen(" << dso_path->c_str() << ") failed.";
        %(logging_function)s << "dlerror() says:\\n" << dlerror();
      }
    }

    if (!module_opened) {
      CloseLibraries(&opened_libraries);
      return false;
    }
  }
""")

# Template to generate code to check if each module initializer correctly
# completed, and cleanup on failures.  This template takes the following
# named parameters.
#   conditional: The conditional expression for successful initialization.
#   uninitializers: The statements needed to uninitialize the modules.
UMBRELLA_INITIALIZER_CHECK_AND_CLEANUP = (
    """  // Check that each module is initialized correctly.
  // Close all previously opened libraries on failure.
  if (%(conditional)s) {
    %(uninitializers)s;
    CloseLibraries(&opened_libraries);
    return false;
  }

  return true;
}
""")

# Template for Initialize, Unininitialize, and IsInitialized functions for each
# module.  This template takes the following named parameters:
#   initialize: Name of the Initialize function.
#   uninitialize: Name of the Uninitialize function.
#   is_initialized: Name of the IsInitialized function.
MODULE_FUNCTION_PROTOTYPES = """bool %(is_initialized)s();
void %(initialize)s(void* module);
void %(uninitialize)s();

"""

# Template for umbrella initializer declaration and associated datatypes.
UMBRELLA_INITIALIZER_PROTOTYPE = (
    """typedef std::map<StubModules, std::vector<std::string> > StubPathMap;

// Umbrella initializer for all the modules in this stub file.
bool InitializeStubs(const StubPathMap& path_map);
""")


def ExtractModuleName(infile_path):
  """Infers the module name from the input file path.

  The input filename is supposed to be in the form "ModuleName.sigs".
  This function splits the filename from the extension on that basename of
  the path and returns that as the module name.

  Args:
    infile_path: String holding the path to the input file.

  Returns:
    The module name as a string.
  """
  basename = os.path.basename(infile_path)

  # This loop continously removes suffixes of the filename separated by a "."
  # character.
  while 1:
    new_basename = os.path.splitext(basename)[0]
    if basename == new_basename:
      break
    else:
      basename = new_basename
  return basename


def ParseSignatures(infile):
  """Parses function signatures in the input file.

  This function parses a file of signatures into a list of dictionaries that
  represent the function signatures in the input file.  Each dictionary has
  the following keys:
    return_type: A string with the return type.
    name: A string with the name of the function.
    params: A list of each function parameter declaration (type + name)

  The format of the input file is one C-style function signature per line, no
  trailing semicolon.  Empty lines are allowed.  An empty line is a line that
  consists purely of whitespace.  Lines that begin with a # or // are considered
  comment lines and are ignored.

  We assume that "int foo(void)" is the same as "int foo()", which is not
  true in C where "int foo()" is equivalent to "int foo(...)".  Our generated
  code is C++, and we do not handle varargs, so this is a case that can be
  ignored for now.

  Args:
    infile: File object holding a text file of function signatures.

  Returns:
    A list of dictionaries, where each dictionary represents one function
    signature.

  Raises:
    BadSignatureError: A line could not be parsed as a signature.
  """
  signatures = []
  for line in infile:
    line = line.strip()
    if line and line[0] != '#' and line[0:2] != '//':
      m = SIGNATURE_REGEX.match(line)
      if m is None:
        raise BadSignatureError('Unparsable line: %s' % line)
      signatures.append(
          {'return_type': m.group('return_type').strip(),
           'name': m.group('name').strip(),
           'params': [arg.strip() for arg in m.group('params').split(',')]})
  return signatures


def WriteWindowsDefFile(module_name, signatures, outfile):
  """Writes a Windows def file to the given output file object.

    The def file format is basically a list of function names.  Generation is
    simple.  After outputting the LIBRARY and EXPORTS lines, print out each
    function name, one to a line, preceded by 2 spaces.

  Args:
    module_name: The name of the module we are writing a stub for.
    signatures: The list of signature hashes, as produced by ParseSignatures,
                to create stubs for.
    outfile: File handle to populate with definitions.
  """
  outfile.write('LIBRARY %s\n' % module_name)
  outfile.write('EXPORTS\n')

  for sig in signatures:
    outfile.write('  %s\n' % sig['name'])


def QuietRun(args, filter=None, write_to=sys.stdout):
  """Invoke |args| as command via subprocess.Popen, filtering lines starting
  with |filter|."""
  popen = subprocess.Popen(args, stdout=subprocess.PIPE)
  out, _ = popen.communicate()
  for line in out.splitlines():
    if not filter or not line.startswith(filter):
      write_to.write(line + '\n')
  return popen.returncode


def CreateWindowsLib(module_name, signatures, intermediate_dir, outdir_path,
                     machine):
  """Creates a Windows library file.

  Calling this function will create a lib file in the outdir_path that exports
  the signatures passed into the object.  A temporary def file will be created
  in the intermediate_dir.

  Args:
    module_name: The name of the module we are writing a stub for.
    signatures: The list of signature hashes, as produced by ParseSignatures,
                to create stubs for.
    intermediate_dir: The directory where the generated .def files should go.
    outdir_path: The directory where generated .lib files should go.
    machine: String holding the machine type, 'X86' or 'X64'.

  Raises:
    SubprocessError: If invoking the Windows "lib" tool fails, this is raised
                     with the error code.
  """
  def_file_path = os.path.join(intermediate_dir,
                               module_name + '.def')
  lib_file_path = os.path.join(outdir_path,
                               module_name + '.lib')
  outfile = open(def_file_path, 'w')
  try:
    WriteWindowsDefFile(module_name, signatures, outfile)
  finally:
    outfile.close()

  # Invoke the "lib" program on Windows to create stub .lib files for the
  # generated definitions.  These .lib files can then be used during
  # delay loading of the dynamic libraries.
  ret = QuietRun(['lib', '/nologo',
                  '/machine:' + machine,
                  '/def:' + def_file_path,
                  '/out:' + lib_file_path],
                 filter='   Creating library')
  if ret != 0:
    raise SubprocessError(
        'Failed creating %s for %s' % (lib_file_path, def_file_path),
        ret)


class PosixStubWriter(object):
  """Creates a file of stub functions for a library that is opened via dlopen.

  Windows provides a function in their compiler known as delay loading, which
  effectively generates a set of stub functions for a dynamic library that
  delays loading of the dynamic library/resolution of the symbols until one of
  the needed functions are accessed.

  In POSIX, RTLD_LAZY does something similar with DSOs.  This is the default
  link mode for DSOs.  However, even though the symbol is not resolved until
  first usage, the DSO must be present at load time of the main binary.

  To simulate the Windows delay load procedure, we need to create a set of
  stub functions that allow for correct linkage of the main binary, but
  dispatch to the dynamically resolved symbol when the module is initialized.

  This class takes a list of function signatures, and generates a set of stub
  functions plus initialization code for them.
  """

  def __init__(self, module_name, export_macro, signatures, logging_function,
               logging_include):
    """Initializes PosixStubWriter for this set of signatures and module_name.

    Args:
      module_name: The name of the module we are writing a stub for.
      export_macro: A preprocessor macro used to annotate stub symbols with
                    an EXPORT marking, to control visibility.
      signatures: The list of signature hashes, as produced by ParseSignatures,
                  to create stubs for.
      logging_function: Function call that will be used for error logging.
      logging_include: Header file where the logging function is defined.
    """
    self.signatures = signatures
    self.module_name = module_name
    self.export_macro = export_macro
    self.logging_function = logging_function
    self.logging_include = logging_include

  @classmethod
  def CStyleIdentifier(cls, identifier):
    """Generates a C style identifier.

    The module_name has all invalid identifier characters removed (anything
    that's not [_a-zA-Z0-9]) and is run through string.capwords to try
    and approximate camel case.

    Args:
      identifier: The string with the module name to turn to C-style.

    Returns:
      A string that can be used as part of a C identifier.
    """
    return string.capwords(re.sub(INVALID_C_IDENT_CHARS, '', identifier))

  @classmethod
  def EnumName(cls, module_name):
    """Gets the enum name for the module.

    Takes the module name and creates a suitable enum name.  The module_name
    is munged to be a valid C identifier then prefixed with the string
    "kModule" to generate a Google style enum name.

    Args:
      module_name: The name of the module to generate an enum name for.

    Returns:
      A string with the name of the enum value representing this module.
    """
    return 'kModule%s' % PosixStubWriter.CStyleIdentifier(module_name)

  @classmethod
  def IsInitializedName(cls, module_name):
    """Gets the name of function that checks initialization of this module.

    The name is in the format IsModuleInitialized.  Where "Module" is replaced
    with the module name, munged to be a valid C identifier.

    Args:
      module_name: The name of the module to generate the function name for.

    Returns:
      A string with the name of the initialization check function.
    """
    return 'Is%sInitialized' % PosixStubWriter.CStyleIdentifier(module_name)

  @classmethod
  def InitializeModuleName(cls, module_name):
    """Gets the name of the function that initializes this module.

    The name is in the format InitializeModule.  Where "Module" is replaced
    with the module name, munged to be a valid C identifier.

    Args:
      module_name: The name of the module to generate the function name for.

    Returns:
      A string with the name of the initialization function.
    """
    return 'Initialize%s' % PosixStubWriter.CStyleIdentifier(module_name)

  @classmethod
  def UninitializeModuleName(cls, module_name):
    """Gets the name of the function that uninitializes this module.

    The name is in the format UninitializeModule.  Where "Module" is replaced
    with the module name, munged to be a valid C identifier.

    Args:
      module_name: The name of the module to generate the function name for.

    Returns:
      A string with the name of the uninitialization function.
    """
    return 'Uninitialize%s' % PosixStubWriter.CStyleIdentifier(module_name)

  @classmethod
  def StubFunctionPointer(cls, signature):
    """Generates a function pointer declaration for the given signature.

    Args:
      signature: A signature hash, as produced by ParseSignatures,
                 representing the function signature.

    Returns:
      A string with the declaration of the function pointer for the signature.
    """
    return 'static %s (*%s_ptr)(%s) = NULL;' % (signature['return_type'],
                                                signature['name'],
                                                ', '.join(signature['params']))

  @classmethod
  def StubFunction(cls, signature):
    """Generates a stub function definition for the given signature.

    The function definitions are created with __attribute__((weak)) so that
    they may be overridden by a real static link or mock versions to be used
    when testing.

    Args:
      signature: A signature hash, as produced by ParseSignatures,
                 representing the function signature.

    Returns:
      A string with the stub function definition.
    """
    return_prefix = ''
    if signature['return_type'] != 'void':
      return_prefix = 'return '

    # Generate the argument list.
    arguments = [re.split('[\*& ]', arg)[-1].strip() for arg in
                 signature['params']]
    # Remove square brackets from arrays, otherwise we will end with a
    # compilation failure.
    for i in range(0, len(arguments)):
      if arguments[i].endswith('[]'):
        arguments[i] = arguments[i][0:-2]

    arg_list = ', '.join(arguments)
    if arg_list == 'void':
      arg_list = ''

    if arg_list != '' and len(arguments) > 1 and arguments[-1] == '...':
      # If the last argument is ... then this is a variadic function.
      if return_prefix != '':
        return VARIADIC_STUB_FUNCTION_DEFINITION % {
            'return_type': signature['return_type'],
            'name': signature['name'],
            'params': ', '.join(signature['params']),
            'arg_list': ', '.join(arguments[0:-1]),
            'last_named_arg': arguments[-2],
            'export': signature.get('export', '')}
      else:
        return VOID_VARIADIC_STUB_FUNCTION_DEFINITION % {
            'name': signature['name'],
            'params': ', '.join(signature['params']),
            'arg_list': ', '.join(arguments[0:-1]),
            'last_named_arg': arguments[-2],
            'export': signature.get('export', '')}
    else:
      # This is a regular function.
      return STUB_FUNCTION_DEFINITION % {
          'return_type': signature['return_type'],
          'name': signature['name'],
          'params': ', '.join(signature['params']),
          'return_prefix': return_prefix,
          'arg_list': arg_list,
          'export': signature.get('export', '')}

  @classmethod
  def WriteImplementationPreamble(cls, header_path, outfile):
    """Write the necessary includes for the implementation file.

    Args:
      header_path: The path to the header file.
      outfile: The file handle to populate.
    """
    outfile.write(IMPLEMENTATION_PREAMBLE % header_path)

  @classmethod
  def WriteUmbrellaInitializer(cls, module_names, namespace, outfile,
                               logging_function):
    """Writes a single function that will open + initialize each module.

    This intializer will take in an stl map of that lists the correct
    dlopen target for each module.  The map type is
    std::map<enum StubModules, vector<std::string>> which matches one module
    to a list of paths to try in dlopen.

    This function is an all-or-nothing function.  If any module fails to load,
    all other modules are dlclosed, and the function returns.  Though it is
    not enforced, this function should only be called once.

    Args:
      module_names: A list with the names of the modules in this stub file.
      namespace: The namespace these functions should be in.
      outfile: The file handle to populate with pointer definitions.
    """
    outfile.write(UMBRELLA_INITIALIZER_START % namespace)
    outfile.write(UMBRELLA_INITIALIZER_CLEANUP_FUNCTION)

    # Create the initialization function that calls all module initializers,
    # checks if they succeeded, and backs out module loads on an error.
    outfile.write(UMBRELLA_INITIALIZER_INITIALIZE_FUNCTION_START % {
        'logging_function': logging_function})
    outfile.write(
        '\n  // Initialize each module if we have not already failed.\n')
    for module in module_names:
      outfile.write('  %s(opened_libraries[%s]);\n' %
                    (PosixStubWriter.InitializeModuleName(module),
                     PosixStubWriter.EnumName(module)))
    outfile.write('\n')

    # Output code to check the initialization status, clean up on error.
    initializer_checks = ['!%s()' % PosixStubWriter.IsInitializedName(name)
                          for name in module_names]
    uninitializers = ['%s()' % PosixStubWriter.UninitializeModuleName(name)
                      for name in module_names]
    outfile.write(UMBRELLA_INITIALIZER_CHECK_AND_CLEANUP % {
        'conditional': ' ||\n      '.join(initializer_checks),
        'uninitializers': ';\n    '.join(uninitializers)})
    outfile.write('\n}  // namespace %s\n' % namespace)

  @classmethod
  def WriteHeaderContents(cls, module_names, namespace, header_guard, outfile,
                          logging_include):
    """Writes a header file for the stub file generated for module_names.

    The header file exposes the following:
       1) An enum, StubModules, listing with an entry for each enum.
       2) A typedef for a StubPathMap allowing for specification of paths to
          search for each module.
       3) The IsInitialized/Initialize/Uninitialize functions for each module.
       4) An umbrella initialize function for all modules.

    Args:
      module_names: A list with the names of each module in this stub file.
      namespace: The namespace these functions should be in.
      header_guard: The macro to use as our header guard.
      outfile: The output handle to populate.
      logging_include: Header file where the logging function is defined.
    """
    outfile.write(STUB_HEADER_PREAMBLE %
                  {'guard_name': header_guard, 'namespace': namespace,
                   'logging_include': logging_include})

    # Generate the Initializer prototypes for each module.
    outfile.write('// Individual module initializer functions.\n')
    for name in module_names:
      outfile.write(MODULE_FUNCTION_PROTOTYPES % {
          'is_initialized': PosixStubWriter.IsInitializedName(name),
          'initialize': PosixStubWriter.InitializeModuleName(name),
          'uninitialize': PosixStubWriter.UninitializeModuleName(name)})

    # Generate the enum for umbrella initializer.
    outfile.write(UMBRELLA_ENUM_START)
    outfile.write('  %s = 0,\n' % PosixStubWriter.EnumName(module_names[0]))
    for name in module_names[1:]:
      outfile.write('  %s,\n' % PosixStubWriter.EnumName(name))
    outfile.write(UMBRELLA_ENUM_END)

    outfile.write(UMBRELLA_INITIALIZER_PROTOTYPE)
    outfile.write(STUB_HEADER_CLOSER % {
        'namespace': namespace, 'guard_name':
        header_guard})

  def WriteImplementationContents(self, namespace, outfile):
    """Given a file handle, write out the stub definitions for this module.

    Args:
      namespace: The namespace these functions should be in.
      outfile: The file handle to populate.
    """
    outfile.write(IMPLEMENTATION_CONTENTS_C_START)
    self.WriteFunctionPointers(outfile)
    self.WriteStubFunctions(outfile)
    outfile.write(IMPLEMENTATION_CONTENTS_C_END)

    outfile.write(NAMESPACE_START % namespace)
    self.WriteModuleInitializeFunctions(outfile)
    outfile.write(NAMESPACE_END % namespace)

  def WriteFunctionPointers(self, outfile):
    """Write the function pointer declarations needed by the stubs.

    We need function pointers to hold the actual location of the function
    implementation returned by dlsym.  This function outputs a pointer
    definition for each signature in the module.

    Pointers will be named with the following pattern "FuntionName_ptr".

    Args:
      outfile: The file handle to populate with pointer definitions.
    """
    outfile.write(FUNCTION_POINTER_SECTION_COMMENT)

    for sig in self.signatures:
      outfile.write('%s\n' % PosixStubWriter.StubFunctionPointer(sig))
    outfile.write('\n')

  def WriteStubFunctions(self, outfile):
    """Write the function stubs to handle dispatching to real implementations.

    Functions that have a return type other than void will look as follows:

      ReturnType FunctionName(A a) {
        return FunctionName_ptr(a);
      }

    Functions with a return type of void will look as follows:

      void FunctionName(A a) {
        FunctionName_ptr(a);
      }

    Args:
      outfile: The file handle to populate.
    """
    outfile.write('// Stubs that dispatch to the real implementations.\n')
    for sig in self.signatures:
      sig['export'] = self.export_macro
      outfile.write('%s\n' % PosixStubWriter.StubFunction(sig))

  def WriteModuleInitializeFunctions(self, outfile):
    """Write functions to initialize/query initlialization of the module.

    This creates 2 functions IsModuleInitialized and InitializeModule where
    "Module" is replaced with the module name, first letter capitalized.

    The InitializeModule function takes a handle that is retrieved from dlopen
    and attempts to assign each function pointer above via dlsym.

    The IsModuleInitialized returns true if none of the required functions
    pointers are NULL.

    Args:
      outfile: The file handle to populate.
    """
    ptr_names = ['%s_ptr' % sig['name'] for sig in self.signatures]

    # Construct the conditional expression to check the initialization of
    # all the function pointers above.  It should generate a conjunction
    # with each pointer on its own line, indented by six spaces to match
    # the indentation level of MODULE_INITIALIZATION_CHECK_FUNCTION.
    initialization_conditional = ' &&\n      '.join(ptr_names)

    outfile.write(MODULE_INITIALIZATION_CHECK_FUNCTION % (
        PosixStubWriter.IsInitializedName(self.module_name),
        initialization_conditional))

    # Create function that initializes the module.
    outfile.write(MODULE_INITIALIZE_START %
                  PosixStubWriter.InitializeModuleName(self.module_name))
    for sig in self.signatures:
      outfile.write(STUB_POINTER_INITIALIZER % {
          'name': sig['name'],
          'return_type': sig['return_type'],
          'parameters': ', '.join(sig['params']),
          'logging_function': self.logging_function})
    outfile.write(MODULE_INITIALIZE_END)

    # Create function that uninitializes the module (sets all pointers to
    # NULL).
    outfile.write(MODULE_UNINITIALIZE_START %
                  PosixStubWriter.UninitializeModuleName(self.module_name))
    for sig in self.signatures:
      outfile.write('  %s_ptr = NULL;\n' % sig['name'])
    outfile.write(MODULE_UNINITIALIZE_END)


def CreateOptionParser():
  """Creates an OptionParser for the configuration options of script.

  Returns:
    A OptionParser object.
  """
  parser = optparse.OptionParser(usage='usage: %prog [options] input')
  parser.add_option('-o',
                    '--output',
                    dest='out_dir',
                    default=None,
                    help='Output location.')
  parser.add_option('-i',
                    '--intermediate_dir',
                    dest='intermediate_dir',
                    default=None,
                    help=('Location of intermediate files. Ignored for %s type'
                          % FILE_TYPE_WIN_DEF))
  parser.add_option('-t',
                    '--type',
                    dest='type',
                    default=None,
                    help=('Type of file. Valid types are "%s" or "%s" or "%s" '
                          'or "%s"' %
                          (FILE_TYPE_POSIX_STUB, FILE_TYPE_WIN_X86,
                           FILE_TYPE_WIN_X64, FILE_TYPE_WIN_DEF)))
  parser.add_option('-s',
                    '--stubfile_name',
                    dest='stubfile_name',
                    default=None,
                    help=('Name of posix_stubs output file. Only valid with '
                          '%s type.' % FILE_TYPE_POSIX_STUB))
  parser.add_option('-p',
                    '--path_from_source',
                    dest='path_from_source',
                    default=None,
                    help=('The relative path from the project root that the '
                          'generated file should consider itself part of (eg. '
                          'third_party/ffmpeg).  This is used to generate the '
                          'header guard and namespace for our initializer '
                          'functions and does NOT affect the physical output '
                          'location of the file like -o does.  Ignored for '
                          '%s and %s types.' %
                          (FILE_TYPE_WIN_X86, FILE_TYPE_WIN_X64)))
  parser.add_option('-e',
                    '--extra_stub_header',
                    dest='extra_stub_header',
                    default=None,
                    help=('File to insert after the system includes in the '
                          'generated stub implementation file. Ignored for '
                          '%s and %s types.' %
                          (FILE_TYPE_WIN_X86, FILE_TYPE_WIN_X64)))
  parser.add_option('-m',
                    '--module_name',
                    dest='module_name',
                    default=None,
                    help=('Name of output DLL or LIB for DEF creation using '
                          '%s type.' % FILE_TYPE_WIN_DEF))
  parser.add_option('-x',
                    '--export_macro',
                    dest='export_macro',
                    default='',
                    help=('A macro to place between the return type and '
                          'function name, e.g. MODULE_EXPORT, to control the '
                          'visibility of the stub functions.'))
  parser.add_option('-l',
                    '--logging-function',
                    dest='logging_function',
                    default='VLOG(1)',
                    help=('Function call that will be used for error logging.'))
  parser.add_option('-n',
                    '--logging-include',
                    dest='logging_include',
                    default='base/logging.h',
                    help=('Header file where the logging function is defined.'))

  return parser


def ParseOptions():
  """Parses the options and terminates program if they are not sane.

  Returns:
    The pair (optparse.OptionValues, [string]), that is the output of
    a successful call to parser.parse_args().
  """
  parser = CreateOptionParser()
  options, args = parser.parse_args()

  if not args:
    parser.error('No inputs specified')

  if options.out_dir is None:
    parser.error('Output location not specified')

  if (options.type not in
      [FILE_TYPE_WIN_X86, FILE_TYPE_WIN_X64, FILE_TYPE_POSIX_STUB,
       FILE_TYPE_WIN_DEF]):
    parser.error('Invalid output file type: %s' % options.type)

  if options.type == FILE_TYPE_POSIX_STUB:
    if options.stubfile_name is None:
      parser.error('Output file name needed for %s' % FILE_TYPE_POSIX_STUB)
    if options.path_from_source is None:
      parser.error('Path from source needed for %s' % FILE_TYPE_POSIX_STUB)

  if options.type == FILE_TYPE_WIN_DEF:
    if options.module_name is None:
      parser.error('Module name needed for %s' % FILE_TYPE_WIN_DEF)

  return options, args


def EnsureDirExists(dir):
  """Creates a directory. Does not use the more obvious 'if not exists: create'
  to avoid race with other invocations of the same code, which will error out
  on makedirs if another invocation has succeeded in creating the directory
  since the existence check."""
  try:
    os.makedirs(dir)
  except:
    if not os.path.isdir(dir):
      raise


def CreateOutputDirectories(options):
  """Creates the intermediate and final output directories.

  Given the parsed options, create the intermediate and final output
  directories if they do not exist.  Returns the paths to both directories
  as a pair.

  Args:
    options: An OptionParser.OptionValues object with the parsed options.

  Returns:
    The pair (out_dir, intermediate_dir), both of which are strings.
  """
  out_dir = os.path.normpath(options.out_dir)
  intermediate_dir = os.path.normpath(options.intermediate_dir)
  if intermediate_dir is None:
    intermediate_dir = out_dir

  EnsureDirExists(out_dir)
  EnsureDirExists(intermediate_dir)

  return out_dir, intermediate_dir


def CreateWindowsLibForSigFiles(sig_files, out_dir, intermediate_dir, machine,
                                export_macro):
  """For each signature file, create a Windows lib.

  Args:
    sig_files: Array of strings with the paths to each signature file.
    out_dir: String holding path to directory where the generated libs go.
    intermediate_dir: String holding path to directory generated intermediate
                      artifacts.
    machine: String holding the machine type, 'X86' or 'X64'.
    export_macro: A preprocessor macro used to annotate stub symbols with
                  an EXPORT marking, to control visibility.
  """
  for input_path in sig_files:
    infile = open(input_path, 'r')
    try:
      signatures = ParseSignatures(infile)
      module_name = ExtractModuleName(os.path.basename(input_path))
      for sig in signatures:
        sig['export'] = export_macro
      CreateWindowsLib(module_name, signatures, intermediate_dir, out_dir,
                       machine)
    finally:
      infile.close()


def CreateWindowsDefForSigFiles(sig_files, out_dir, module_name):
  """For all signature files, create a single Windows def file.

  Args:
    sig_files: Array of strings with the paths to each signature file.
    out_dir: String holding path to directory where the generated def goes.
    module_name: Name of the output DLL or LIB which will link in the def file.
  """
  signatures = []
  for input_path in sig_files:
    infile = open(input_path, 'r')
    try:
      signatures += ParseSignatures(infile)
    finally:
      infile.close()

  def_file_path = os.path.join(
      out_dir, os.path.splitext(os.path.basename(module_name))[0] + '.def')
  outfile = open(def_file_path, 'w')

  try:
    WriteWindowsDefFile(module_name, signatures, outfile)
  finally:
    outfile.close()


def CreatePosixStubsForSigFiles(sig_files, stub_name, out_dir,
                                intermediate_dir, path_from_source,
                                extra_stub_header, export_macro,
                                logging_function, logging_include):
  """Create a POSIX stub library with a module for each signature file.

  Args:
    sig_files: Array of strings with the paths to each signature file.
    stub_name: String with the basename of the generated stub file.
    out_dir: String holding path to directory for the .h files.
    intermediate_dir: String holding path to directory for the .cc files.
    path_from_source: String with relative path of generated files from the
                      project root.
    extra_stub_header: String with path to file of extra lines to insert
                       into the generated header for the stub library.
    export_macro: A preprocessor macro used to annotate stub symbols with
                  an EXPORT marking, to control visibility.
    logging_function: Function call that will be used for error logging.
    logging_include: Header file where the logging function is defined.
  """
  header_base_name = stub_name + '.h'
  header_path = os.path.join(out_dir, header_base_name)
  impl_path = os.path.join(intermediate_dir, stub_name + '.cc')

  module_names = [ExtractModuleName(path) for path in sig_files]
  namespace = path_from_source.replace('/', '_').lower()
  header_guard = '%s_' % namespace.upper()
  header_include_path = os.path.join(path_from_source, header_base_name)

  # First create the implementation file.
  impl_file = open(impl_path, 'w')
  try:
    # Open the file, and create the preamble which consists of a file
    # header plus any necessary includes.
    PosixStubWriter.WriteImplementationPreamble(header_include_path,
                                                impl_file)
    if extra_stub_header is not None:
      extra_header_file = open(extra_stub_header, 'r')
      try:
        impl_file.write('\n')
        for line in extra_header_file:
          impl_file.write(line)
        impl_file.write('\n')
      finally:
        extra_header_file.close()

    # For each signature file, generate the stub population functions
    # for that file.  Each file represents one module.
    for input_path in sig_files:
      name = ExtractModuleName(input_path)
      infile = open(input_path, 'r')
      try:
        signatures = ParseSignatures(infile)
      finally:
        infile.close()
      writer = PosixStubWriter(name, export_macro, signatures, logging_function,
                               logging_include)
      writer.WriteImplementationContents(namespace, impl_file)

    # Lastly, output the umbrella function for the file.
    PosixStubWriter.WriteUmbrellaInitializer(module_names, namespace,
                                             impl_file, logging_function)
  finally:
    impl_file.close()

  # Then create the associated header file.
  header_file = open(header_path, 'w')
  try:
    PosixStubWriter.WriteHeaderContents(module_names, namespace,
                                        header_guard, header_file,
                                        logging_include)
  finally:
    header_file.close()


def main():
  options, args = ParseOptions()
  out_dir, intermediate_dir = CreateOutputDirectories(options)

  if options.type == FILE_TYPE_WIN_X86:
    CreateWindowsLibForSigFiles(args, out_dir, intermediate_dir, 'X86',
                                options.export_macro)
  elif options.type == FILE_TYPE_WIN_X64:
    CreateWindowsLibForSigFiles(args, out_dir, intermediate_dir, 'X64',
                                options.export_macro)
  elif options.type == FILE_TYPE_POSIX_STUB:
    CreatePosixStubsForSigFiles(args, options.stubfile_name, out_dir,
                                intermediate_dir, options.path_from_source,
                                options.extra_stub_header, options.export_macro,
                                options.logging_function,
                                options.logging_include)
  elif options.type == FILE_TYPE_WIN_DEF:
    CreateWindowsDefForSigFiles(args, out_dir, options.module_name)


if __name__ == '__main__':
  main()



# Copyright (c) 2017 The WebRTC project authors. All Rights Reserved.
#
# Use of this source code is governed by a BSD-style license
# that can be found in the LICENSE file in the root of the source
# tree. An additional intellectual property rights grant can be found
# in the file PATENTS.  All contributing project authors may
# be found in the AUTHORS file in the root of the source tree.
"""EvaluationScore factory class.
"""

import logging

from . import exceptions
from . import eval_scores


class EvaluationScoreWorkerFactory(object):
    """Factory class used to instantiate evaluation score workers.

  The ctor gets the parametrs that are used to instatiate the evaluation score
  workers.
  """

    def __init__(self, polqa_tool_bin_path, echo_metric_tool_bin_path):
        self._score_filename_prefix = None
        self._polqa_tool_bin_path = polqa_tool_bin_path
        self._echo_metric_tool_bin_path = echo_metric_tool_bin_path

    def SetScoreFilenamePrefix(self, prefix):
        self._score_filename_prefix = prefix

    def GetInstance(self, evaluation_score_class):
        """Creates an EvaluationScore instance given a class object.

    Args:
      evaluation_score_class: EvaluationScore class object (not an instance).

    Returns:
      An EvaluationScore instance.
    """
        if self._score_filename_prefix is None:
            raise exceptions.InitializationException(
                'The score file name prefix for evaluation score workers is not set'
            )
        logging.debug('factory producing a %s evaluation score',
                      evaluation_score_class)

        if evaluation_score_class == eval_scores.PolqaScore:
            return eval_scores.PolqaScore(self._score_filename_prefix,
                                          self._polqa_tool_bin_path)
        elif evaluation_score_class == eval_scores.EchoMetric:
            return eval_scores.EchoMetric(self._score_filename_prefix,
                                          self._echo_metric_tool_bin_path)
        else:
            return evaluation_score_class(self._score_filename_prefix)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
MAX_ARGS = 15

boilerplate = "/* This Source Code Form is subject to the terms of the Mozilla Public\n\
 * License, v. 2.0. If a copy of the MPL was not distributed with this\n\
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */\n"

includes = "#include \"RefCounted.h\"\n"

refcountclass = "class RefCountTaskWrapper : public gmp_args_base {\n\
public:\n\
  RefCountTaskWrapper(GMPTask* aTask, RefCounted* aRefCounted)\n\
    : mTask(aTask)\n\
    , mRefCounted(aRefCounted)\n\
  {}\n\
  virtual void Run() {\n\
    mTask->Run();\n\
  }\n\
  virtual void Destroy() {\n\
    mTask->Destroy();\n\
    gmp_args_base::Destroy();\n\
  }\n\
private:\n\
  ~RefCountTaskWrapper() {}\n\
\n\
  GMPTask* mTask;\n\
  RefPtr<RefCounted> mRefCounted;\n\
};\n"


def gen_args_type(args, member):
    if member:
        ret = ["C o"]
    else:
        ret = []
    ret.append("M m")
    for arg in range(0, args):
        ret.append("A%d a%d"%(arg, arg))
    return ", ".join(ret)

def gen_args(args, member):
    if member:
        ret = ["o"]
    else:
        ret = []
    ret.append("m")
    for arg in range(0, args):
        ret.append("a%d"%(arg))
    return ", ".join(ret)

def gen_args_(args):
    ret = []
    for arg in range(0, args):
        ret.append("a%d_"%(arg))
    return ", ".join(ret)

def gen_init(args, r = False, member = False):
    if member:
        ret = ["o_ (o)"]
    else:
        ret = []
    ret.append("m_ (m)")

    if r:
        ret.append("r_ (r)")

    for arg in range(0, args):
        ret.append("a%d_ (a%d)"%(arg, arg))
    return ", ".join(ret)

def gen_typenames(args, member):
    if member:
        ret = ["typename C"]
    else:
        ret = []
    ret.append("typename M")

    for arg in range(0, args):
        ret.append("typename A%d"%(arg))
    return ", ".join(ret)

def gen_types(args, member):
    if member:
        ret = ["C"]
    else:
        ret = []
    ret.append("M")
    for arg in range(0, args):
        ret.append("A%d"%(arg))
    return ", ".join(ret)


def generate_class_template(args, ret = False, member = True):
    print "// %d arguments --"%args
    if member:
        nm = "m"
    else:
        nm = "nm"

    if not ret:
        print "template<"+ gen_typenames(args, member) + "> class gmp_args_%s_%d : public gmp_args_base {"%(nm, args)
    else:
        print "template<"+ gen_typenames(args, member) + ", typename R> class gmp_args_%s_%d_ret : public gmp_args_base {"%(nm, args)

    print " public:"

    if not ret:
        print "  gmp_args_%s_%d ("%(nm, args) + gen_args_type(args, member) + ") :"
        print "    " + gen_init(args, False, member) + "  {}"
    else:
        print "  gmp_args_%s_%d_ret ("%(nm, args) + gen_args_type(args, member) + ", R* r) :"
        print "    " + gen_init(args, True, member) + "  {}"
        print "  virtual bool returns_value() const {\n    return true;\n  }"
    print
    print "  void Run() {"
    if ret:
        print "    *r_ =",
    else:
        print "   ",
    if member:
        print "((*o_).*m_) (" + gen_args_(args) + ");"
    else:
        print "m_ (" + gen_args_(args) + ");"
    print "  }"
    print
    print " private:"
    if member:
        print "  C o_;"
    print "  M m_;"
    if ret:
        print "  R* r_;"
    for arg in range(0, args):
        print "  A%d a%d_;"%(arg, arg)
    print "};"
    print
    print
    print

def generate_function_template(args, member):
    if member:
        nm = "m"
        NM = "";
    else:
        nm = "nm"
        NM = "NM";

    print "// %d arguments --"%args
    print "template<" + gen_typenames(args, member) + ">"
    print "gmp_args_%s_%d<"%(nm, args) + gen_types(args, member) + ">* WrapTask%s ("%NM + gen_args_type(args, member) + ") {"
    print "  return new gmp_args_%s_%d<"%(nm, args) + gen_types(args, member) + ">"
    print "    (" + gen_args(args, member) + ");"
    print "}"
    print
    if member:
        print "template<" + gen_typenames(args, member) + ">"
        print "GMPTask*"
        print "WrapTaskRefCounted%s ("%NM + gen_args_type(args, member) + ") {"
        print "  GMPTask *t = WrapTask%s ("%NM + gen_args(args, member) + ");"
        print "  return new RefCountTaskWrapper(t, o);"
        print "}"
        print

def generate_function_template_ret(args, member):
    if member:
        nm = "m"
        NM = "";
    else:
        nm = "nm"
        NM = "NM";
    print "// %d arguments --"%args
    print "template<" + gen_typenames(args, member) + ", typename R>"
    print "gmp_args_%s_%d_ret<"%(nm, args) + gen_types(args, member) + ", R>* WrapTask%sRet ("%NM + gen_args_type(args, member) + ", R* r) {"
    print "  return new gmp_args_%s_%d_ret<"%(nm, args) + gen_types(args, member) + ", R>"
    print "    (" + gen_args(args, member) + ", r);"
    print "}"
    print



print boilerplate
print
print includes
print
print refcountclass
print

for num_args in range (0, MAX_ARGS):
    generate_class_template(num_args, False, False)
    generate_class_template(num_args, True, False)
    generate_class_template(num_args, False, True)
    generate_class_template(num_args, True, True)


print
print
print

for num_args in range(0, MAX_ARGS):
    generate_function_template(num_args, False)
    generate_function_template_ret(num_args, False)
    generate_function_template(num_args, True)
    generate_function_template_ret(num_args, True)






# Copyright (c) 2017 The WebRTC project authors. All Rights Reserved.
#
# Use of this source code is governed by a BSD-style license
# that can be found in the LICENSE file in the root of the source
# tree. An additional intellectual property rights grant can be found
# in the file PATENTS.  All contributing project authors may
# be found in the AUTHORS file in the root of the source tree.
"""Signal processing utility module.
"""

import array
import logging
import os
import sys
import enum

try:
    import numpy as np
except ImportError:
    logging.critical('Cannot import the third-party Python package numpy')
    sys.exit(1)

try:
    import pydub
    import pydub.generators
except ImportError:
    logging.critical('Cannot import the third-party Python package pydub')
    sys.exit(1)

try:
    import scipy.signal
    import scipy.fftpack
except ImportError:
    logging.critical('Cannot import the third-party Python package scipy')
    sys.exit(1)

from . import exceptions


class SignalProcessingUtils(object):
    """Collection of signal processing utilities.
  """

    @enum.unique
    class MixPadding(enum.Enum):
        NO_PADDING = 0
        ZERO_PADDING = 1
        LOOP = 2

    def __init__(self):
        pass

    @classmethod
    def LoadWav(cls, filepath, channels=1):
        """Loads wav file.

    Args:
      filepath: path to the wav audio track file to load.
      channels: number of channels (downmixing to mono by default).

    Returns:
      AudioSegment instance.
    """
        if not os.path.exists(filepath):
            logging.error('cannot find the <%s> audio track file', filepath)
            raise exceptions.FileNotFoundError()
        return pydub.AudioSegment.from_file(filepath,
                                            format='wav',
                                            channels=channels)

    @classmethod
    def SaveWav(cls, output_filepath, signal):
        """Saves wav file.

    Args:
      output_filepath: path to the wav audio track file to save.
      signal: AudioSegment instance.
    """
        return signal.export(output_filepath, format='wav')

    @classmethod
    def CountSamples(cls, signal):
        """Number of samples per channel.

    Args:
      signal: AudioSegment instance.

    Returns:
      An integer.
    """
        number_of_samples = len(signal.get_array_of_samples())
        assert signal.channels > 0
        assert number_of_samples % signal.channels == 0
        return number_of_samples / signal.channels

    @classmethod
    def GenerateSilence(cls, duration=1000, sample_rate=48000):
        """Generates silence.

    This method can also be used to create a template AudioSegment instance.
    A template can then be used with other Generate*() methods accepting an
    AudioSegment instance as argument.

    Args:
      duration: duration in ms.
      sample_rate: sample rate.

    Returns:
      AudioSegment instance.
    """
        return pydub.AudioSegment.silent(duration, sample_rate)

    @classmethod
    def GeneratePureTone(cls, template, frequency=440.0):
        """Generates a pure tone.

    The pure tone is generated with the same duration and in the same format of
    the given template signal.

    Args:
      template: AudioSegment instance.
      frequency: Frequency of the pure tone in Hz.

    Return:
      AudioSegment instance.
    """
        if frequency > template.frame_rate >> 1:
            raise exceptions.SignalProcessingException('Invalid frequency')

        generator = pydub.generators.Sine(sample_rate=template.frame_rate,
                                          bit_depth=template.sample_width * 8,
                                          freq=frequency)

        return generator.to_audio_segment(duration=len(template), volume=0.0)

    @classmethod
    def GenerateWhiteNoise(cls, template):
        """Generates white noise.

    The white noise is generated with the same duration and in the same format
    of the given template signal.

    Args:
      template: AudioSegment instance.

    Return:
      AudioSegment instance.
    """
        generator = pydub.generators.WhiteNoise(
            sample_rate=template.frame_rate,
            bit_depth=template.sample_width * 8)
        return generator.to_audio_segment(duration=len(template), volume=0.0)

    @classmethod
    def AudioSegmentToRawData(cls, signal):
        samples = signal.get_array_of_samples()
        if samples.typecode != 'h':
            raise exceptions.SignalProcessingException(
                'Unsupported samples type')
        return np.array(signal.get_array_of_samples(), np.int16)

    @classmethod
    def Fft(cls, signal, normalize=True):
        if signal.channels != 1:
            raise NotImplementedError('multiple-channel FFT not implemented')
        x = cls.AudioSegmentToRawData(signal).astype(np.float32)
        if normalize:
            x /= max(abs(np.max(x)), 1.0)
        y = scipy.fftpack.fft(x)
        return y[:len(y) / 2]

    @classmethod
    def DetectHardClipping(cls, signal, threshold=2):
        """Detects hard clipping.

    Hard clipping is simply detected by counting samples that touch either the
    lower or upper bound too many times in a row (according to `threshold`).
    The presence of a single sequence of samples meeting such property is enough
    to label the signal as hard clipped.

    Args:
      signal: AudioSegment instance.
      threshold: minimum number of samples at full-scale in a row.

    Returns:
      True if hard clipping is detect, False otherwise.
    """
        if signal.channels != 1:
            raise NotImplementedError(
                'multiple-channel clipping not implemented')
        if signal.sample_width != 2:  # Note that signal.sample_width is in bytes.
            raise exceptions.SignalProcessingException(
                'hard-clipping detection only supported for 16 bit samples')
        samples = cls.AudioSegmentToRawData(signal)

        # Detect adjacent clipped samples.
        samples_type_info = np.iinfo(samples.dtype)
        mask_min = samples == samples_type_info.min
        mask_max = samples == samples_type_info.max

        def HasLongSequence(vector, min_legth=threshold):
            """Returns True if there are one or more long sequences of True flags."""
            seq_length = 0
            for b in vector:
                seq_length = seq_length + 1 if b else 0
                if seq_length >= min_legth:
                    return True
            return False

        return HasLongSequence(mask_min) or HasLongSequence(mask_max)

    @classmethod
    def ApplyImpulseResponse(cls, signal, impulse_response):
        """Applies an impulse response to a signal.

    Args:
      signal: AudioSegment instance.
      impulse_response: list or numpy vector of float values.

    Returns:
      AudioSegment instance.
    """
        # Get samples.
        assert signal.channels == 1, (
            'multiple-channel recordings not supported')
        samples = signal.get_array_of_samples()

        # Convolve.
        logging.info(
            'applying %d order impulse response to a signal lasting %d ms',
            len(impulse_response), len(signal))
        convolved_samples = scipy.signal.fftconvolve(in1=samples,
                                                     in2=impulse_response,
                                                     mode='full').astype(
                                                         np.int16)
        logging.info('convolution computed')

        # Cast.
        convolved_samples = array.array(signal.array_type, convolved_samples)

        # Verify.
        logging.debug('signal length: %d samples', len(samples))
        logging.debug('convolved signal length: %d samples',
                      len(convolved_samples))
        assert len(convolved_samples) > len(samples)

        # Generate convolved signal AudioSegment instance.
        convolved_signal = pydub.AudioSegment(data=convolved_samples,
                                              metadata={
                                                  'sample_width':
                                                  signal.sample_width,
                                                  'frame_rate':
                                                  signal.frame_rate,
                                                  'frame_width':
                                                  signal.frame_width,
                                                  'channels': signal.channels,
                                              })
        assert len(convolved_signal) > len(signal)

        return convolved_signal

    @classmethod
    def Normalize(cls, signal):
        """Normalizes a signal.

    Args:
      signal: AudioSegment instance.

    Returns:
      An AudioSegment instance.
    """
        return signal.apply_gain(-signal.max_dBFS)

    @classmethod
    def Copy(cls, signal):
        """Makes a copy os a signal.

    Args:
      signal: AudioSegment instance.

    Returns:
      An AudioSegment instance.
    """
        return pydub.AudioSegment(data=signal.get_array_of_samples(),
                                  metadata={
                                      'sample_width': signal.sample_width,
                                      'frame_rate': signal.frame_rate,
                                      'frame_width': signal.frame_width,
                                      'channels': signal.channels,
                                  })

    @classmethod
    def MixSignals(cls,
                   signal,
                   noise,
                   target_snr=0.0,
                   pad_noise=MixPadding.NO_PADDING):
        """Mixes `signal` and `noise` with a target SNR.

    Mix `signal` and `noise` with a desired SNR by scaling `noise`.
    If the target SNR is +/- infinite, a copy of signal/noise is returned.
    If `signal` is shorter than `noise`, the length of the mix equals that of
    `signal`. Otherwise, the mix length depends on whether padding is applied.
    When padding is not applied, that is `pad_noise` is set to NO_PADDING
    (default), the mix length equals that of `noise` - i.e., `signal` is
    truncated. Otherwise, `noise` is extended and the resulting mix has the same
    length of `signal`.

    Args:
      signal: AudioSegment instance (signal).
      noise: AudioSegment instance (noise).
      target_snr: float, numpy.Inf or -numpy.Inf (dB).
      pad_noise: SignalProcessingUtils.MixPadding, default: NO_PADDING.

    Returns:
      An AudioSegment instance.
    """
        # Handle infinite target SNR.
        if target_snr == -np.Inf:
            # Return a copy of noise.
            logging.warning('SNR = -Inf, returning noise')
            return cls.Copy(noise)
        elif target_snr == np.Inf:
            # Return a copy of signal.
            logging.warning('SNR = +Inf, returning signal')
            return cls.Copy(signal)

        # Check signal and noise power.
        signal_power = float(signal.dBFS)
        noise_power = float(noise.dBFS)
        if signal_power == -np.Inf:
            logging.error('signal has -Inf power, cannot mix')
            raise exceptions.SignalProcessingException(
                'cannot mix a signal with -Inf power')
        if noise_power == -np.Inf:
            logging.error('noise has -Inf power, cannot mix')
            raise exceptions.SignalProcessingException(
                'cannot mix a signal with -Inf power')

        # Mix.
        gain_db = signal_power - noise_power - target_snr
        signal_duration = len(signal)
        noise_duration = len(noise)
        if signal_duration <= noise_duration:
            # Ignore `pad_noise`, `noise` is truncated if longer that `signal`, the
            # mix will have the same length of `signal`.
            return signal.overlay(noise.apply_gain(gain_db))
        elif pad_noise == cls.MixPadding.NO_PADDING:
            # `signal` is longer than `noise`, but no padding is applied to `noise`.
            # Truncate `signal`.
            return noise.overlay(signal, gain_during_overlay=gain_db)
        elif pad_noise == cls.MixPadding.ZERO_PADDING:
            # TODO(alessiob): Check that this works as expected.
            return signal.overlay(noise.apply_gain(gain_db))
        elif pad_noise == cls.MixPadding.LOOP:
            # `signal` is longer than `noise`, extend `noise` by looping.
            return signal.overlay(noise.apply_gain(gain_db), loop=True)
        else:
            raise exceptions.SignalProcessingException('invalid padding type')



# Copyright (c) 2017 The WebRTC project authors. All Rights Reserved.
#
# Use of this source code is governed by a BSD-style license
# that can be found in the LICENSE file in the root of the source
# tree. An additional intellectual property rights grant can be found
# in the file PATENTS.  All contributing project authors may
# be found in the AUTHORS file in the root of the source tree.
"""Unit tests for the simulation module.
"""

import logging
import os
import shutil
import tempfile
import unittest

import mock
import pydub

from . import audioproc_wrapper
from . import eval_scores_factory
from . import evaluation
from . import external_vad
from . import signal_processing
from . import simulation
from . import test_data_generation_factory


class TestApmModuleSimulator(unittest.TestCase):
    """Unit tests for the ApmModuleSimulator class.
  """

    def setUp(self):
        """Create temporary folders and fake audio track."""
        self._output_path = tempfile.mkdtemp()
        self._tmp_path = tempfile.mkdtemp()

        silence = pydub.AudioSegment.silent(duration=1000, frame_rate=48000)
        fake_signal = signal_processing.SignalProcessingUtils.GenerateWhiteNoise(
            silence)
        self._fake_audio_track_path = os.path.join(self._output_path,
                                                   'fake.wav')
        signal_processing.SignalProcessingUtils.SaveWav(
            self._fake_audio_track_path, fake_signal)

    def tearDown(self):
        """Recursively delete temporary folders."""
        shutil.rmtree(self._output_path)
        shutil.rmtree(self._tmp_path)

    def testSimulation(self):
        # Instance dependencies to mock and inject.
        ap_wrapper = audioproc_wrapper.AudioProcWrapper(
            audioproc_wrapper.AudioProcWrapper.DEFAULT_APM_SIMULATOR_BIN_PATH)
        evaluator = evaluation.ApmModuleEvaluator()
        ap_wrapper.Run = mock.MagicMock(name='Run')
        evaluator.Run = mock.MagicMock(name='Run')

        # Instance non-mocked dependencies.
        test_data_generator_factory = (
            test_data_generation_factory.TestDataGeneratorFactory(
                aechen_ir_database_path='',
                noise_tracks_path='',
                copy_with_identity=False))
        evaluation_score_factory = eval_scores_factory.EvaluationScoreWorkerFactory(
            polqa_tool_bin_path=os.path.join(os.path.dirname(__file__),
                                             'fake_polqa'),
            echo_metric_tool_bin_path=None)

        # Instance simulator.
        simulator = simulation.ApmModuleSimulator(
            test_data_generator_factory=test_data_generator_factory,
            evaluation_score_factory=evaluation_score_factory,
            ap_wrapper=ap_wrapper,
            evaluator=evaluator,
            external_vads={
                'fake':
                external_vad.ExternalVad(
                    os.path.join(os.path.dirname(__file__),
                                 'fake_external_vad.py'), 'fake')
            })

        # What to simulate.
        config_files = ['apm_configs/default.json']
        input_files = [self._fake_audio_track_path]
        test_data_generators = ['identity', 'white_noise']
        eval_scores = ['audio_level_mean', 'polqa']

        # Run all simulations.
        simulator.Run(config_filepaths=config_files,
                      capture_input_filepaths=input_files,
                      test_data_generator_names=test_data_generators,
                      eval_score_names=eval_scores,
                      output_dir=self._output_path)

        # Check.
        # TODO(alessiob): Once the TestDataGenerator classes can be configured by
        # the client code (e.g., number of SNR pairs for the white noise test data
        # generator), the exact number of calls to ap_wrapper.Run and evaluator.Run
        # is known; use that with assertEqual.
        min_number_of_simulations = len(config_files) * len(input_files) * len(
            test_data_generators)
        self.assertGreaterEqual(len(ap_wrapper.Run.call_args_list),
                                min_number_of_simulations)
        self.assertGreaterEqual(len(evaluator.Run.call_args_list),
                                min_number_of_simulations)

    def testInputSignalCreation(self):
        # Instance simulator.
        simulator = simulation.ApmModuleSimulator(
            test_data_generator_factory=(
                test_data_generation_factory.TestDataGeneratorFactory(
                    aechen_ir_database_path='',
                    noise_tracks_path='',
                    copy_with_identity=False)),
            evaluation_score_factory=(
                eval_scores_factory.EvaluationScoreWorkerFactory(
                    polqa_tool_bin_path=os.path.join(os.path.dirname(__file__),
                                                     'fake_polqa'),
                    echo_metric_tool_bin_path=None)),
            ap_wrapper=audioproc_wrapper.AudioProcWrapper(
                audioproc_wrapper.AudioProcWrapper.
                DEFAULT_APM_SIMULATOR_BIN_PATH),
            evaluator=evaluation.ApmModuleEvaluator())

        # Inexistent input files to be silently created.
        input_files = [
            os.path.join(self._tmp_path, 'pure_tone-440_1000.wav'),
            os.path.join(self._tmp_path, 'pure_tone-1000_500.wav'),
        ]
        self.assertFalse(
            any([os.path.exists(input_file) for input_file in (input_files)]))

        # The input files are created during the simulation.
        simulator.Run(config_filepaths=['apm_configs/default.json'],
                      capture_input_filepaths=input_files,
                      test_data_generator_names=['identity'],
                      eval_score_names=['audio_level_peak'],
                      output_dir=self._output_path)
        self.assertTrue(
            all([os.path.exists(input_file) for input_file in (input_files)]))

    def testPureToneGenerationWithTotalHarmonicDistorsion(self):
        logging.warning = mock.MagicMock(name='warning')

        # Instance simulator.
        simulator = simulation.ApmModuleSimulator(
            test_data_generator_factory=(
                test_data_generation_factory.TestDataGeneratorFactory(
                    aechen_ir_database_path='',
                    noise_tracks_path='',
                    copy_with_identity=False)),
            evaluation_score_factory=(
                eval_scores_factory.EvaluationScoreWorkerFactory(
                    polqa_tool_bin_path=os.path.join(os.path.dirname(__file__),
                                                     'fake_polqa'),
                    echo_metric_tool_bin_path=None)),
            ap_wrapper=audioproc_wrapper.AudioProcWrapper(
                audioproc_wrapper.AudioProcWrapper.
                DEFAULT_APM_SIMULATOR_BIN_PATH),
            evaluator=evaluation.ApmModuleEvaluator())

        # What to simulate.
        config_files = ['apm_configs/default.json']
        input_files = [os.path.join(self._tmp_path, 'pure_tone-440_1000.wav')]
        eval_scores = ['thd']

        # Should work.
        simulator.Run(config_filepaths=config_files,
                      capture_input_filepaths=input_files,
                      test_data_generator_names=['identity'],
                      eval_score_names=eval_scores,
                      output_dir=self._output_path)
        self.assertFalse(logging.warning.called)

        # Warning expected.
        simulator.Run(
            config_filepaths=config_files,
            capture_input_filepaths=input_files,
            test_data_generator_names=['white_noise'],  # Not allowed with THD.
            eval_score_names=eval_scores,
            output_dir=self._output_path)
        logging.warning.assert_called_with('the evaluation failed: %s', (
            'The THD score cannot be used with any test data generator other than '
            '"identity"'))

    #   # Init.
    #   generator = test_data_generation.IdentityTestDataGenerator('tmp')
    #   input_signal_filepath = os.path.join(
    #       self._test_data_cache_path, 'pure_tone-440_1000.wav')

    #   # Check that the input signal is generated.
    #   self.assertFalse(os.path.exists(input_signal_filepath))
    #   generator.Generate(
    #       input_signal_filepath=input_signal_filepath,
    #       test_data_cache_path=self._test_data_cache_path,
    #       base_output_path=self._base_output_path)
    #   self.assertTrue(os.path.exists(input_signal_filepath))

    #   # Check input signal properties.
    #   input_signal = signal_processing.SignalProcessingUtils.LoadWav(
    #       input_signal_filepath)
    #   self.assertEqual(1000, len(input_signal))
